#!/usr/bin/env python3
"""Helper program to simulate various subprocess behaviors."""

import os
import logging
from pathlib import Path
import signal
import sys
import time

logger = logging.getLogger('subprocess_helper')


def main(args):
    out = sys.stdout
    for arg in args:
        if arg == 'err:':
            out = sys.stderr
        elif arg == 'out:':
            out = sys.stdout
        elif arg == 'in:':
            print(sys.stdin.readline().rstrip(), file=out)
        elif arg == 'cwd:':
            print(Path.cwd(), file=out)
        elif arg.startswith('env:'):
            print(os.environ[arg[4:]], file=out)
        elif arg.startswith('sleep:'):
            time.sleep(float(arg[6:]))
        elif arg.startswith('touch:'):
            Path(arg[6:]).touch()
        elif arg.startswith('log:'):
            logger.error(arg[4:])
        elif arg.startswith('ignore:'):
            signal.signal(getattr(signal, arg[7:]), signal.SIG_IGN)
        elif arg.startswith('exit:'):
            return int(arg[5:])
        else:
            print(arg, file=out)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



import asyncio
import contextlib
from functools import partial
import itertools
import logging
import os
from pathlib import Path
import pytest
import signal
from subprocess import CalledProcessError
import sys
import time
import traceback

from asyncjobs import logcontext
from verify_events import EventVerifier, ExpectedJobEvents, Whatever

logger = logging.getLogger(__name__)


@contextlib.asynccontextmanager
async def abort_in(when=None, assert_on_escape=True):
    """Simulate Ctrl+C (aka. KeyboardInterrupt/SIGINT) after the given delay.

    This will schedule a SIGINT signal in the current process after the given
    number of seconds if we're still within the context. Upon context exit,
    the signal is cancelled.

    If the signal is handled within the context, no further action is taken.
    Otherwise, if the signal/KeyboardInterrupt escapes the context, an error
    is logged and (unless assert_on_escape is explicitly disabled) an
    assertion is raised (to fail the current test).
    """
    if when is None:  # Do nothing
        yield
        return

    async def wait_and_kill():
        await asyncio.sleep(when)
        logger.warning('Raising SIGINT to simulate Ctrl+C…')
        os.kill(os.getpid(), signal.SIGINT)

    task = asyncio.create_task(wait_and_kill())
    try:
        yield
    except KeyboardInterrupt:
        logger.error('SIGINT/KeyboardInterrupt escaped the context!')
        if assert_on_escape:
            assert False, 'SIGINT/KeyboardInterrupt escaped the context!'
    finally:
        if not task.done():
            logger.debug(f'context complete before {when}')
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task


@contextlib.contextmanager
def assert_elapsed_time(predicate):
    """Measure time used in context and assert that predicate holds.

    Calculate elapsed time in context, and assert predicate(elapsed).
    """
    before = time.time()
    try:
        yield
    finally:
        after = time.time()
        logger.debug(f'Passing {after - before} to predicate {predicate}')
        assert predicate(after - before)


class TBasicJob:
    """Async jobs with test instrumentation."""

    def __init__(
        self,
        name,
        deps=None,
        *,
        before=None,
        async_sleep=0,
        coro=None,
        spawn=None,
        await_spawn=False,
        result=None,
    ):
        self.name = name
        self.deps = deps
        self.before = set() if before is None else set(before)
        self.async_sleep = async_sleep
        self.coro = coro
        self.spawn = [] if spawn is None else spawn
        self.await_spawn = await_spawn
        self.result = f'{name} done' if result is None else result

        self.xevents = ExpectedJobEvents(name)
        self.xevents.add('add')
        self.xevents.add('start')

        # self.deps is processed _before_ .__call__() is invoked
        if self.deps:
            if 'MISSING' in self.deps:
                self.xevents.add('finish', fate='failed')  # expect KeyError
            else:
                self.xevents.add(
                    'await results', jobs=list(self.deps), pending=Whatever
                )

    def descendants(self):
        for spawn in self.spawn:
            yield spawn
            yield from spawn.descendants()

    async def __call__(self, ctx):
        # Await results from dependencies
        dep_results = ctx.deps
        if self.deps:
            self.xevents.add('awaited results')
        logger.debug(f'Results from deps: {dep_results}')

        # Sleep, if requested
        if self.async_sleep:
            logger.info(f'Async sleep for {self.async_sleep} seconds…')
            await asyncio.sleep(self.async_sleep)
            logger.info('Finished async sleep')

        # Do the scheduled work
        if self.coro is not None:
            try:
                ctx.tjob = self  # give test code access to test job instance
                coro = self.coro(ctx)
                assert asyncio.iscoroutine(coro), f'not a coroutine: {coro}'
                self.result = await coro
            except asyncio.CancelledError:
                logger.info('Cancelled!')
                self.xevents.add('finish', fate='cancelled')
                raise
            except Exception as e:
                logger.info(f'Raising exception: {e!r}')
                traceback.print_exc()
                self.xevents.add('finish', fate='failed')
                raise

        # Spawn children jobs
        for job in self.spawn:
            ctx.add_job(job.name, job, job.deps)

        if self.await_spawn and self.spawn:
            spawn = [job.name for job in self.spawn]
            self.xevents.add('await results', jobs=spawn, pending=spawn)
            await ctx.results(*[job.name for job in self.spawn])
            self.xevents.add('awaited results')

        # Verify that our dependents are still waiting for us
        for b in self.before:
            assert b in ctx._scheduler.tasks  # The other job has started
            assert not ctx._scheduler.tasks[b].done()  # but not yet finished

        # Return result or raise exception
        if isinstance(self.result, Exception):
            logger.info(f'Raising exception: {self.result!r}')
            self.xevents.add('finish', fate='failed')
            raise self.result
        else:
            logger.info(f'Returning result: {self.result!r}')
            self.xevents.add('finish', fate='success')
            return self.result


class TExternalWorkJob(TBasicJob):
    """Test jobs with thread/subprocess capabilities."""

    def __init__(self, *args, thread=None, argv=None, **kwargs):
        super().__init__(*args, **kwargs)
        if thread:
            assert argv is None and self.coro is None
            self.coro = partial(self.run_thread, thread)
        elif argv:
            assert thread is None and self.coro is None
            self.coro = partial(self.run_subprocess, argv)

    async def run_thread(self, thread, ctx, **kwargs):
        with self.thread_xevents():
            return await ctx.call_in_thread(thread, ctx, **kwargs)

    async def run_subprocess(self, argv, ctx):
        with self.subprocess_xevents(argv, may_cancel=True):
            return await ctx.run_in_subprocess(argv, check=True)

    @contextlib.contextmanager
    def thread_xevents(self, reuse_ticket=False):
        if not reuse_ticket:
            self.xevents.add('await worker slot')
            self.xevents.add('awaited worker slot', may_cancel=True)
        self.xevents.add(
            'start work in thread', func=Whatever, may_cancel=True
        )
        try:
            yield
            self.xevents.add('finish work in thread', fate='success')
        except asyncio.CancelledError:
            self.xevents.add(
                'finish work in thread', fate='cancelled', may_cancel=True
            )
            raise
        except Exception:
            self.xevents.add('finish work in thread', fate='failed')
            raise

    @contextlib.contextmanager
    def subprocess_xevents(
        self, argv, result=None, may_cancel=False, reuse_ticket=False
    ):
        if not reuse_ticket:
            self.xevents.add('await worker slot')
            self.xevents.add('awaited worker slot', may_cancel=may_cancel)
        self.xevents.add(
            'start work in subprocess', argv=argv, may_cancel=may_cancel
        )

        try:
            yield
            if result is None:  # Normal return: assume exit code 0
                self.xevents.add('finish work in subprocess', returncode=0)
        except asyncio.CancelledError:
            if result is None:  # Assume cancellation was expected
                self.xevents.add(
                    'subprocess terminate',
                    argv=argv,
                    pid=Whatever,
                    may_cancel=may_cancel,
                )
                self.xevents.add(
                    'finish work in subprocess',
                    returncode=-15,
                    may_cancel=may_cancel,
                )
            raise
        except Exception as e:
            if result is None and isinstance(e, CalledProcessError):
                self.xevents.add(
                    'finish work in subprocess', returncode=e.returncode
                )
            raise
        finally:
            if result == 'terminate' or result == 'kill':
                self.xevents.add(
                    'subprocess terminate',
                    argv=argv,
                    pid=Whatever,
                    may_cancel=may_cancel,
                )
                if result == 'kill':
                    self.xevents.add(
                        'subprocess kill',
                        argv=argv,
                        pid=Whatever,
                        may_cancel=may_cancel,
                    )
                    result = -9
                else:
                    result = -15
            if result is not None:
                self.xevents.add(
                    'finish work in subprocess',
                    returncode=result,
                    may_cancel=may_cancel,
                )


@pytest.fixture(params=[1, 2, 4, 100])
def num_workers(request):
    return request.param


@pytest.fixture
def num_jobs(num_workers):
    """How many jobs to run in stress tests."""
    return min(100, num_workers * 10)


def verify_number_of_tasks(n):
    all_tasks = asyncio.all_tasks()
    if len(asyncio.all_tasks()) != n:
        message = f'There are {len(all_tasks)} != {n} running!'
        logger.error(message)
        for task in all_tasks:
            logger.error(f'  - {task}')
        raise RuntimeError(message)


@pytest.fixture
def scheduler_with_workers(num_workers):
    def make_scheduler_class(*bases):
        class _Scheduler(*bases):
            pass

        logger.info(f'creating scheduler with {num_workers} worker threads')
        return partial(_Scheduler, workers=num_workers)

    return make_scheduler_class


def mock_argv(*args):
    mock_path = Path(__file__).parent / 'subprocess_helper.py'
    return [sys.executable, '-u', str(mock_path)] + list(args)


@contextlib.contextmanager
def adjusted_logger_level(logger, level):
    old_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(old_level)


@contextlib.contextmanager
def verified_events(scheduler, todo):
    """Check events generated by scheduler against expectations.

    Verify that the given scheduler generates the expected events while in
    this context. We set the scheduler's event_handler to collect all events
    that occur while in the context, and upon exiting the context, we verify
    the events emitted by the scheduler against the events expected by the
    TBasicJob instances in 'todo'.
    """
    initial_xevents = [j.xevents for j in todo]
    spawned_xevents = [
        j.xevents for j in itertools.chain(*[j.descendants() for j in todo])
    ]
    ev = EventVerifier()
    scheduler.event_handler = ev
    yield
    ev.verify_all(initial_xevents, spawned_xevents)


# Used to signal the expectation of a cancelled task
Cancelled = object()


def verify_tasks(tasks, expects):
    errors = 0
    NotGiven = object()

    def fail(job_name, expect, actual=NotGiven):
        nonlocal errors
        if actual is NotGiven:
            logger.error(f'{job_name}: {expect}')
        else:
            logger.error(f'{job_name}: expected {expect!r}, actual {actual!r}')
        errors += 1

    for name in set(tasks.keys()) | set(expects.keys()):
        try:
            expect = expects[name]
            task = tasks[name]
        except KeyError:
            e = 'present' if name in expects else 'missing'
            t = 'present' if name in tasks else 'missing'
            fail(name, f'{e} in expects, {t} in tasks')
            continue
        if expect is Cancelled:
            if not task.cancelled():
                fail(name, Cancelled, task)
        elif isinstance(expect, Exception):
            e = task.exception()
            if not isinstance(e, expect.__class__):
                fail(name, expect.__class__, type(e))
            if e.args != expect.args:
                fail(name, expect.args, e.args)
        elif callable(expect):  # verify by passing task to predicate
            if not expect(task):
                fail(name, f'{expect}({task}) failed')
        else:
            if task.result() != expect:
                fail(name, expect, task.result())
    return errors == 0


class ListHandler(logging.Handler):
    """A log handler implementation that simply stores log records in a list.

    Used to test that log records are handled/formatted correctly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.records = []
        self.messages = []

    def emit(self, record):
        self.records.append(record)
        self.messages.append(self.format(record))


@pytest.fixture
def logger_with_listhandler():
    logger = logging.getLogger('test')
    handler = ListHandler()
    handler.setFormatter(logcontext.Formatter())
    logger.addHandler(handler)
    return logger, handler


@pytest.fixture
def verify_output(capfd, logger_with_listhandler):
    def _verify_one_output(expect_lines_from_streams, actual_text):
        actual_lines = actual_text.split('\n')
        logger.debug(
            f'Verifying {actual_lines} against {expect_lines_from_streams}'
        )
        assert actual_lines.pop() == ''  # last line was terminated
        for actual_line in actual_lines:
            # Expect this line is the next line from one of our streams
            expected = set()
            for expect_lines in expect_lines_from_streams:
                if expect_lines and actual_line == expect_lines[0]:  # found it
                    expect_lines.pop(0)
                    break
                elif expect_lines:
                    expected.add(expect_lines[0])
            else:  # no match for any stream
                assert False, f'actual: {actual_line!r}, expect: {expected!r}'
        # no more lines expected
        for e in expect_lines_from_streams:
            if len(e):
                assert False, f'expected lines not found: {e!r}'

    def _verify_output(
        expect_stdout_streams,
        expect_stderr_streams=None,
        expect_log_streams=None,
    ):
        actual = capfd.readouterr()
        _verify_one_output(expect_stdout_streams, actual.out)
        if expect_stderr_streams is not None:
            _verify_one_output(expect_stderr_streams, actual.err)
        if expect_log_streams is not None:
            _, listhandler = logger_with_listhandler
            actual_log = ''.join(msg + '\n' for msg in listhandler.messages)
            _verify_one_output(expect_log_streams, actual_log)
        return True

    return _verify_output



import asyncio
import logging
import sys

from asyncjobs import stream_mux, decorated_work

from conftest import adjusted_logger_level, verify_tasks


async def job_coro(ctx):
    with ctx.decoration(
        decorate_out=f'{ctx.name}/out: ', decorate_err=f'{ctx.name}/ERR: ',
    ):
        with ctx.stdout as outf, ctx.stderr as errf:
            print('FOO', file=outf)
            print('BAR', file=errf)
            return 123


async def noop():
    pass


# These tests verify that the StreamMuxes used by decorated_work are hooked up
# to the same event loop as the scheduler itself. Regressions here are not
# reproducible when everything runs inside the same asyncio test framework,
# hence these tests drive asyncio.run directly.


def test_instantiating_logmux_before_asyncio_run_is_ok():
    stream_mux.StreamMux(sys.stdout)
    asyncio.run(noop())


def test_instantiating_logmux_after_asyncio_run_is_ok():
    asyncio.run(noop())
    stream_mux.StreamMux(sys.stdout)


def test_one_simple_job_through_default_muxes(verify_output, num_workers):
    scheduler = decorated_work.Scheduler(workers=num_workers)
    scheduler.add_job('foo', job_coro)
    # Prevent stream_mux DEBUG messages on stderr
    with adjusted_logger_level(stream_mux.logger, logging.INFO):
        done = asyncio.run(scheduler.run())
    assert verify_tasks(done, {'foo': 123})
    assert verify_output([['foo/out: FOO']], [['foo/ERR: BAR']])


def test_one_simple_job_through_custom_muxes(verify_output, num_workers):
    scheduler = decorated_work.Scheduler(
        workers=num_workers,
        outmux=stream_mux.StreamMux(sys.stderr),
        errmux=stream_mux.StreamMux(sys.stdout),
    )
    scheduler.add_job('foo', job_coro)
    # Prevent stream_mux DEBUG messages on stderr
    with adjusted_logger_level(stream_mux.logger, logging.INFO):
        done = asyncio.run(scheduler.run())
    assert verify_tasks(done, {'foo': 123})
    assert verify_output([['foo/ERR: BAR']], [['foo/out: FOO']])



"""Basic job scheduler.

A 'job' is a _coroutine_ with a _name_, and a set of _dependencies_:

Coroutine

Jobs are executed in an asynchronous context. As usual in async programming,
a coroutine should never directly invoke any long, blocking operations, as
this affects the entire async framework and prevents other jobs from
progressing. Such work must instead be 'await'ed. See the external_work module
for an extension to this module which provides helpers for starting and
awaiting synchronous work in worker threads and/or subprocesses.

A job coroutine may finish in one of three ways:
  - complete successfully, by returning a value.
  - fail, by raising an exception which aborts the coroutine.
  - get cancelled, in which case this framework raises a CancelledError inside
    the coroutine, causing it to abort.

Name

Jobs have a name that is unique to the Scheduler instance.
The name is used in several contexts:
  - When jobs specify dependency relationships between eachother.
  - In events emitted while Scheduler.run() is executing
  - As a key in the job results dict returned from Scheduler.run()

Dependencies

Jobs may specify a set of dependencies which are names of other jobs that must
finish executing successfully before this job can proceed. The results of
those other jobs are made available in the Context object passed to the job
coroutine. If any of the dependencies fail (either by exception, or by getting
cancelled), then this job will also be cancelled.

Scheduler

Jobs are created by calling Scheduler.add_job(). When the Scheduler.run() is
called, each of the added jobs/coroutines will be invoked once with a Context
object (see below) and wrapped in an asyncio.Task
object which starts it running concurrently with other coroutines.

The Scheduler.run() ends when all jobs have finished (one way or another) and
returns a dict mapping job names to their corresponding asyncio.Future objects
that encapsulate their result (return value, exception, or cancellation).

This module implements the most basic Scheduler. Other modules in this package
extend the Scheduler class (and accompanying Context class) with more features.

Context

Each job's coroutine is called with a single argument: ctx, which is a Context
object the job may use to interact with the Scheduler or other parts of the
surrounding framework. For example:

  - While running, a job may await the result of one or more other jobs by
    passing their names to ctx.results() (this is also how dependencies between
    jobs are implemented): A job 'foo' may depend on jobs 'bar' and 'baz' by
    awaiting their results:

        results = ctx.results('bar', 'baz')

  - A job may spawn another job by calling ctx.add_job().
"""
import asyncio
import concurrent.futures
import logging
import time

logger = logging.getLogger(__name__)


class Context:
    """Helper class instantiated per job and passed to the job coroutine.

    This is the interface available to the job coroutine for communicating
    with the Scheduler, and other parts of the surrounding scheduler
    environment.
    """

    def __init__(self, name, scheduler):
        self.name = name
        self.deps = None  # set by the coroutine wrapper in Scheduler.add_job()
        self._scheduler = scheduler

    def event(self, event, **kwargs):
        """Emit a scheduling event from this job."""
        self._scheduler.event(event, job=self.name, **kwargs)

    async def results(self, *job_names):
        """Wait until the given jobs are finished, and return their results.

        Returns a dictionary mapping job names to their results.

        Raise KeyError if any of the given jobs have not been added to the
        scheduler. If any of the given jobs fail (either by raising an
        exception or by being cancelled) then raise CancelledError here to
        cancel this job as a consequence.
        """
        assert self._scheduler.running
        tasks = [self._scheduler.tasks[n] for n in job_names]
        pending = [n for n, t in zip(job_names, tasks) if not t.done()]
        self.event('await results', jobs=list(job_names), pending=pending)
        if pending:
            logger.debug(f'{self.name} waiting for {", ".join(pending)}…')
        try:
            results = dict(zip(job_names, await asyncio.gather(*tasks)))
        except Exception:
            logger.info(f'{self.name} cancelled due to failed dependency')
            raise asyncio.CancelledError
        self.event('awaited results')
        logger.debug(f'{self.name} returning {results} from .results()')
        return results

    def add_job(self, name, coro, deps=None):
        """Add a job to be run.

        Schedule the job to start immediately. Raise ValueError if a job with
        the same name has already been added.
        """
        self._scheduler.add_job(name, coro, deps)


class Scheduler:
    """Async job scheduler. Run job coroutines as concurrent tasks.

    Jobs are added by passing their names and coroutines to .add_job(). The
    coroutines will be scheduled for execution when .run() is called. While
    running, jobs may use the Context object passed to the coroutine to await
    each other's results or spawn new jobs. When all jobs are completed
    (successfully or otherwise), .run() will return a dictionary with all the
    job results.
    """

    def __init__(self, *, event_handler=None, context_class=Context):
        self.jobs = {}  # job name -> coroutine
        self.tasks = {}  # job name -> Task object, aka. (future) job result
        self.running = False
        self.event_handler = event_handler

        assert issubclass(context_class, Context)
        self.context_class = context_class

    def __contains__(self, job_name):
        return job_name in self.jobs

    def event(self, event, **kwargs):
        if self.event_handler is None:
            return
        d = {'timestamp': time.time(), 'event': event}
        d.update(kwargs)
        logger.debug(f'Posting event: {d}')
        self.event_handler(d)

    @staticmethod
    def _fate(future):
        """Return a word describing the state of the given future."""
        if not future.done():
            return 'unfinished'
        elif future.cancelled():
            return 'cancelled'
        elif future.exception() is not None:
            logger.warning(f'FAILED future: {future}:')
            # raise future.exception()
            return 'failed'
        else:
            return 'success'

    def _start_job(self, name):
        assert name in self.jobs
        assert name not in self.tasks

        ctx = self.context_class(name, self)
        task = asyncio.create_task(self.jobs[name](ctx), name=name)
        self.tasks[name] = task
        self.event('start', job=name)
        task.add_done_callback(
            lambda task: self.event('finish', job=name, fate=self._fate(task))
        )

    def add_job(self, name, coro, deps=None):
        """Add a job (aka. named coroutine) to be run.

        If 'deps' is given, it must be a set of names of other jobs that will
        be awaited before coro is started. The results of those jobs is made
        available in the ctx.deps dict.

        If we're already running (i.e. inside .run()) schedule the job
        immediately, otherwise the job will be scheduled when .run() is called.
        Raise ValueError if a job with the same name has already been added.
        """
        if name in self.jobs:
            raise ValueError(f'Cannot add job named {name}. Already added!')

        logger.debug(f'Adding job named {name} with deps={deps!r}')
        if deps is not None:

            async def deps_before_coro(ctx, wrapped_coro=coro):
                assert ctx.deps is None
                logger.debug(f'Awaiting dependencies: {deps}…')
                ctx.deps = await ctx.results(*deps)
                return await wrapped_coro(ctx)

            coro = deps_before_coro

        self.jobs[name] = coro
        self.event('add', job=name)
        if self.running:
            self._start_job(name)

    async def _run_tasks(self, return_when):
        logger.info('Waiting for all jobs to complete…')
        self.event('await tasks', jobs=list(self.tasks.keys()))
        shutting_down = False
        while any(not t.done() for t in self.tasks.values()):
            # NEVER exit this loop while there are tasks still running.
            try:
                # Await the tasks that are currently running. More tasks may be
                # spawned while we're waiting, and those are not awaited here.
                await asyncio.wait(
                    self.tasks.values(), return_when=return_when
                )
                # We know _some_ task completed, successfully or not
                assert any(t.done() for t in self.tasks.values())

                if not shutting_down:
                    # It is time to shut down, yet?
                    if return_when == concurrent.futures.FIRST_COMPLETED:
                        shutting_down = True
                    elif return_when == concurrent.futures.FIRST_EXCEPTION:
                        shutting_down = any(
                            t.done() and (t.cancelled() or t.exception())
                            for t in self.tasks.values()
                        )
                    else:
                        assert return_when == concurrent.futures.ALL_COMPLETED
                        shutting_down = all(
                            t.done() for t in self.tasks.values()
                        )
            except BaseException as e:
                logger.warning(f'{self.__class__.__name__} aborted by {e!r}')
                shutting_down = True
            finally:
                if shutting_down:
                    # Keep cancelling tasks until all are finished
                    logger.info('Shutting down…')
                    self.event('cancelling tasks')
                    return_when = concurrent.futures.ALL_COMPLETED
                    for name, task in self.tasks.items():
                        if not task.done():
                            logger.info(f'Cancelling {name}…')
                            task.cancel()
        self.event('awaited tasks')

    async def run(self, *, keep_going=False):
        """Run until all jobs are finished.

        If keep_going is disabled (the default), the first failing job (i.e.
        a job that raises an exception) will cause us to cancel all other
        concurrent and remaining jobs and return as soon as possible.

        If keep_going is enabled, we will keep running as long as there are
        jobs to do. Only the jobs that depend (directly or indirectly) on the
        failing job(s) will be cancelled.

        Return a dictionary mapping job names to the corresponding
        asyncio.Future objects that encapsulate their result (return value,
        exception, or cancellation).
        """
        logger.debug('Running…')
        self.event('start', keep_going=keep_going, num_jobs=len(self.jobs))
        if keep_going:
            return_when = concurrent.futures.ALL_COMPLETED
        else:
            return_when = concurrent.futures.FIRST_EXCEPTION

        if self.jobs:
            for name in self.jobs.keys():
                self._start_job(name)

            self.running = True
            await self._run_tasks(return_when)
            self.running = False
        else:
            logger.warning('Nothing to do!')
        self.event('finish', num_tasks=len(self.tasks))
        return self.tasks



import asyncio
import pytest

from asyncjobs.stream_mux import StreamMux

pytestmark = pytest.mark.asyncio


async def test_no_output(verify_output):
    with StreamMux():
        pass
    assert verify_output([])


async def test_one_stream_undecorated(verify_output):
    with StreamMux() as mux, mux.new_stream() as f:
        print('This is the first line', file=f)
        print('This is the second line', file=f)
    assert verify_output(
        [['This is the first line', 'This is the second line']]
    )


async def test_two_streams_undecorated(verify_output):
    with StreamMux() as mux, mux.new_stream() as f:
        print('This is stream 1 line 1', file=f)
        with mux.new_stream() as g:
            print('This is stream 2 line 1', file=g)
            print('This is stream 2 line 2', file=g)
        print('This is stream 1 line 2', file=f)
    assert verify_output(
        [
            ['This is stream 1 line 1', 'This is stream 1 line 2'],
            ['This is stream 2 line 1', 'This is stream 2 line 2'],
        ],
    )


async def test_one_stream_decorated(verify_output):
    with StreamMux() as mux:
        decorator = StreamMux.simple_decorator('[pre]{}[post]')  # long-winded
        with mux.new_stream(decorator) as f:
            print('This is the first line', file=f)
            print('This is the second line', file=f)
    assert verify_output(
        [
            [
                '[pre]This is the first line[post]',
                '[pre]This is the second line[post]',
            ]
        ]
    )


async def test_two_streams_decorated(verify_output):
    with StreamMux() as mux:
        with mux.new_stream(b'1>>{}<<1') as f:  # shorter version
            print('This is stream 1 line 1', file=f)
            with mux.new_stream('2>>{}<<2') as g:
                print('This is stream 2 line 1', file=g)
                print('This is stream 2 line 2', file=g)
            print('This is stream 1 line 2', file=f)
    assert verify_output(
        [
            ['1>>This is stream 1 line 1<<1', '1>>This is stream 1 line 2<<1'],
            ['2>>This is stream 2 line 1<<2', '2>>This is stream 2 line 2<<2'],
        ],
    )


async def test_one_charwise_stream_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    with StreamMux() as mux, mux.new_stream('<{}>') as f:
        for c in s:
            f.write(c)
    assert verify_output([['<foo>', '<bar>', '<baz>']])


async def test_one_charwise_interrupted_stream_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    with StreamMux() as mux, mux.new_stream('<{}>') as f:
        for c in s:
            f.write(c)
            f.flush()
            await asyncio.sleep(0.001)
    assert verify_output([['<foo>', '<bar>', '<baz>']])


async def test_two_charwise_streams_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    t = '123\n456\n789'
    with StreamMux() as mux:
        with mux.new_stream(b'<{}>') as f, mux.new_stream('[{}]') as g:
            for c, d in zip(s, t):
                f.write(c)
                g.write(d)
    assert verify_output(
        [['<foo>', '<bar>', '<baz>'], ['[123]', '[456]', '[789]']]
    )


async def test_two_charwise_interrupted_streams_decorated(verify_output):
    s = 'foo\nbar\nbaz'
    t = '123\n456\n789'
    with StreamMux() as mux:
        with mux.new_stream(b'<{}>') as f, mux.new_stream('[{}]') as g:
            for c, d in zip(s, t):
                f.write(c)
                g.write(d)
                f.flush()
                g.flush()
                await asyncio.sleep(0.001)
    assert verify_output(
        [['<foo>', '<bar>', '<baz>'], ['[123]', '[456]', '[789]']]
    )


async def test_one_bytewise_stream_with_garbage(capfdbinary):
    lines = [
        b'first line...',
        b'latin-1: \xc6\xd8\xc5...',
        b'utf-8:   \xe2\x9c\x94\xe2\x88\x80\xe2\x9c\x98...',
        b'f8 - ff: \xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff...',
        b'last line without newline',
    ]
    bytestring = b'\n'.join(lines)
    prefix, suffix = '❰'.encode('utf-8'), '❱\n'.encode('utf-8')
    expect_bytestring = b''.join(prefix + line + suffix for line in lines)
    with StreamMux() as mux, mux.new_stream('❰{}❱') as f:
        f.buffer.write(bytestring)
    actual = capfdbinary.readouterr()
    assert actual.out == expect_bytestring
    assert actual.err == b''


async def test_one_bytewise_stream_in_binary_mode_with_garbage(capfdbinary):
    lines = [
        b'first line...',
        b'latin-1: \xc6\xd8\xc5...',
        b'utf-8:   \xe2\x9c\x94\xe2\x88\x80\xe2\x9c\x98...',
        b'f8 - ff: \xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff...',
        b'last line without newline',
    ]
    bytestring = b'\n'.join(lines)
    prefix = b'>>> '  # test passing bytes w/o placeholder to simple_decorator
    expect_bytestring = b''.join(prefix + line + b'\n' for line in lines)
    with StreamMux() as mux, mux.new_stream(prefix, mode='wb') as f:
        f.write(bytestring)
    actual = capfdbinary.readouterr()
    assert actual.out == expect_bytestring
    assert actual.err == b''


async def test_write_to_file(tmp_path):
    path = tmp_path / 'file'
    with path.open('w') as f, StreamMux(f) as mux, mux.new_stream() as g:
        g.write('first line\n')
        await asyncio.sleep(0.001)
        g.write('second line\n')

    assert path.read_text() == 'first line\nsecond line\n'


async def test_follow_file_reads_from_beginning(tmp_path, verify_output):
    path = tmp_path / 'file'
    with path.open('w') as f, StreamMux() as mux:
        print('first line', file=f, flush=True)
        with mux.follow_file(path, '<{}>'):
            print('second line', file=f, flush=True)
        print('third line', file=f, flush=True)  # not watched
    assert verify_output([['<first line>', '<second line>']])


async def test_follow_file_first_read_is_immediate(tmp_path, verify_output):
    path = tmp_path / 'file'
    with path.open('w') as f, StreamMux() as mux:
        print('first line', file=f, flush=True)
        with mux.follow_file(path, '<{}>'):
            await asyncio.sleep(0.001)
            assert verify_output([['<first line>']])
    assert verify_output([])


async def test_follow_file_second_read_after_period(tmp_path, verify_output):
    path = tmp_path / 'file'
    with path.open('w') as f, StreamMux() as mux:
        print('first line', file=f, flush=True)
        with mux.follow_file(path, '<{}>', period=0.01):
            await asyncio.sleep(0.001)
            assert verify_output([['<first line>']])
            print('second line', file=f, flush=True)
            await asyncio.sleep(0.001)
            assert verify_output([])
            await asyncio.sleep(0.01)
            assert verify_output([['<second line>']])
    assert verify_output([])


async def test_follow_file_read_after_writer_close(tmp_path, verify_output):
    path = tmp_path / 'file'
    f = path.open('w')
    print('first line', file=f, flush=True)
    with StreamMux() as mux, mux.follow_file(path, '<{}>', period=0.01):
        await asyncio.sleep(0.001)
        assert verify_output([['<first line>']])
        print('second line', file=f, flush=True)
        f.close()
        await asyncio.sleep(0.01)
        assert verify_output([['<second line>']])
    assert verify_output([])


async def test_follow_file_ignores_short_rewrites(tmp_path, verify_output):
    path = tmp_path / 'file'
    f = path.open('w')
    print('this is the first line', file=f, flush=True)
    with StreamMux() as mux, mux.follow_file(path, '<{}>', period=0.01):
        await asyncio.sleep(0.001)
        assert verify_output([['<this is the first line>']])
        print('second line', file=f, flush=True)
        f.close()
        f = path.open('w')
        print('short rewrite', file=f, flush=True)  # not seen by reader
        f.close()
    assert verify_output([])  # misses both 'second line' and 'short rewrite'


async def test_follow_file_reads_appends(tmp_path, verify_output):
    path = tmp_path / 'file'
    f = path.open('w')
    print('first line', file=f, flush=True)
    with StreamMux() as mux, mux.follow_file(path, '<{}>', period=0.01):
        await asyncio.sleep(0.001)
        assert verify_output([['<first line>']])
        print('second line', file=f, flush=True)
        f.close()
        f = path.open('a')
        print('third line', file=f, flush=True)
        f.close()
    assert verify_output([['<second line>', '<third line>']])


async def test_internal_errors_are_propagated(tmp_path):
    path = tmp_path / 'file'
    f = path.open('w')
    with pytest.raises(ValueError):
        with StreamMux(f) as mux, mux.new_stream() as g:
            g.write('first line\n')
            g.flush()
            await asyncio.sleep(0.001)
            f.flush()
            f.close()
            g.write('second line\n')  # raises ValueError: write to closed file
        # context exit raises ValueError: flush of closed file

    assert path.read_text() == 'first line\n'



import contextlib
import functools
import logging
import sys

from . import external_work, logcontext, stream_mux

logger = logging.getLogger(__name__)


def decorated_job(decorate_out=None, decorate_err=None, decorate_log=None):
    """Setup stdout/stderr/log decoration for the given coroutine.

    This is the same as wrapping the entire coroutine in:

        with ctx.decoration(...):
            ...

    Use as a decorator:

        @decorated_job()
        async def my_job(ctx):
            logger.info('hello')
            await asyncio.sleep(1)
            logger.info(world')

        @decorated_job(decorate_out='DECORATED: ')
        async def my_other_job(ctx):
            with ctx.stdout as f:
                print('hello', file=f)
                await asyncio.sleep(1)
                print(again', file=f)

    See Context.decoration() for more details
    """

    def wrap(coro):
        @functools.wraps(coro)
        async def wrapped_coro(ctx):
            with ctx.decoration(
                decorate_out=decorate_out,
                decorate_err=decorate_err,
                decorate_log=decorate_log,
            ):
                return await coro(ctx)

        return wrapped_coro

    return wrap


class Context(external_work.Context):
    """API for jobs that want to decorate their logs/stdout/stderr.

    This enables jobs to create stdout/stderr streams that are automatically
    decorated and multiplexed onto the stdout/stderr streams controlled by the
    below Scheduler.

    Redirection of the actual stdout/stderr file descriptors is automatically
    done for subprocesses. Log statements (made in the job coroutine or its
    callees, or from threads run by the job) are not redirected, but are
    decorated nonetheless (as long as logcontext.Formatter is used as the log
    formatter).

    Other output (from the job coroutine, or from thread workers) can be
    decorated by writing into the file descriptor retrieved from
    self.stdout.open() or self.stderr.open() (or by using self.stdout or
    self.stderr as context managers).
    """

    def __init__(self, *args):
        super().__init__(*args)
        self.stdout = None
        self.stderr = None

    @contextlib.contextmanager
    def decoration(
        self, *, decorate_out=None, decorate_err=None, decorate_log=None,
    ):
        """Setup context with decorated stdout/stderr streams + decorated log.

        Inside this context, self.stdout and self.stderr are DecoratedStream
        objects (decorated by 'decorate_out' and 'decorate_err', respectively)
        that may be opened/entered in order to write decorated output.

        Likewise, 'decorate_log' is set as the log decorator for the current
        context (the current job).

        If 'decorate_log' is not given, but 'decorate_err' is given,
        'decorate_log' will default to 'decorate_err'. Otherwise, all
        decorators default to None, i.e. no decoration.

        Note that the output streams (self.stdout, self.stderr) are used as the
        default stdout/stderr stream for subprocesses started within this
        context (see .subprocess() below). Similarly, 'decorate_log' is
        automatically applied to the log context of threads started within
        this context (see .call_in_thread() below).
        """
        if decorate_log is True:
            decorate_log = decorate_err
        assert self.stdout is None and self.stderr is None
        try:
            self.stdout = self._scheduler.outmux.new_stream(decorate_out)
            self.stderr = self._scheduler.errmux.new_stream(decorate_err)
            with logcontext.Decorator.use(decorate_log):
                yield
        finally:
            if self.stdout is not None:
                self.stdout.close()
                self.stdout = None
            if self.stderr is not None:
                self.stderr.close()
                self.stderr = None

    @contextlib.contextmanager
    def follow_file(self, path, use_stderr=False, period=0.1):
        """Echo the given file to self.stdout (or self.stderr).

        This sets up a StreamMux poller to periodically check for new content
        and echo lines from the given file. By default, the self.stdout mux
        (and decorator) is used, but pass use_stderr=True to use the
        self.stderr mux (and decorator) instead.

        This will only work inside the above .decoration() context, otherwise
        a RuntimeError is raised.

        Lines read from 'path' will be passed through 'decorator' before being
        written to this StreamMux' shared output.
        """
        stream = self.stderr if use_stderr else self.stdout
        if stream is None:
            raise RuntimeError('Cannot follow file outside decoration context')
        with stream.mux.follow_file(path, stream.decorator, period=period):
            yield

    async def call_in_thread(self, func, *args, **kwargs):
        """Apply the current log context (if any) to threaded function call."""
        decorator = logcontext.Decorator.get(logcontext.current_context())
        assert decorator is None or callable(decorator)

        @functools.wraps(func)
        def wrapper(*args):
            with logcontext.Decorator.use(decorator):
                return func(*args)

        return await super().call_in_thread(wrapper, *args, **kwargs)

    @contextlib.asynccontextmanager
    async def subprocess(
        self, argv, stdout=None, stderr=None, ticket=None, **kwargs
    ):
        """Pass redirected out/err as stdout/stderr to the subprocess.

        Only if stdout/stderr is not already customized by the caller.
        """
        async with self._use_ticket(ticket, only_reserve=True) as ticket:
            with contextlib.ExitStack() as stack:
                if stdout is None and self.stdout is not None:
                    stdout = stack.enter_context(self.stdout)
                if stderr is None and self.stderr is not None:
                    stderr = stack.enter_context(self.stderr)
                async with super().subprocess(
                    argv, stdout=stdout, stderr=stderr, ticket=ticket, **kwargs
                ) as proc:
                    yield proc


class Scheduler(external_work.Scheduler):
    """Run jobs with automatic decoration of their stdout/stderr + logs.

    Extend Scheduler with two StreamMux instances - one for stdout and one for
    stderr - to be used by job coroutines for redirecting (and decorating)
    their stdout/stderr through to a single pair of output streams controlled
    by this scheduler instance.
    """

    def __init__(
        self, *, outmux=None, errmux=None, context_class=Context, **kwargs,
    ):
        self.outmux = outmux
        self.errmux = errmux

        assert issubclass(context_class, Context)
        super().__init__(context_class=context_class, **kwargs)

    async def _run_tasks(self, *args, **kwargs):
        if self.outmux is None:
            self.outmux = stream_mux.StreamMux(sys.stdout)
        if self.errmux is None:
            self.errmux = stream_mux.StreamMux(sys.stderr)

        logger.debug('Starting StreamMux instances…')
        with self.outmux, self.errmux:
            await super()._run_tasks(*args, **kwargs)



import asyncio
import concurrent.futures
import pytest

from asyncjobs import logcontext


def run_in_no_context(func, *args, **kwargs):
    return func(*args, **kwargs)


def run_in_async_context(func, *args, **kwargs):
    async def coro():
        return func(*args, **kwargs)

    return asyncio.run(coro())


def run_in_thread_context(func, *args, **kwargs):
    with concurrent.futures.ThreadPoolExecutor() as executor:
        future = executor.submit(func, *args, **kwargs)
        return future.result()


@pytest.fixture(
    params=[run_in_no_context, run_in_async_context, run_in_thread_context]
)
def run_in_any_context(request):
    return request.param


class TestFormatterInAnyContext:
    @pytest.fixture(autouse=True)
    def setup(self, run_in_any_context, logger_with_listhandler):
        self.logger, self.handler = logger_with_listhandler
        self.handler.setFormatter(logcontext.Formatter('[%(message)s]'))
        self.run_in_context = run_in_any_context

    def test_no_decorators_does_not_decorate(self):
        self.run_in_context(self.logger.error, 'FOO')
        assert self.handler.messages == ['[FOO]']

    def test_unrelated_decorator_does_not_decorate(self):
        with logcontext.Decorator.use(lambda line: f'>>>{line}<<<', 12345):
            self.run_in_context(self.logger.error, 'FOO')
        assert self.handler.messages == ['[FOO]']

    def test_default_decorator_only_in_no_context(self):
        def in_context():
            with logcontext.Decorator.use(lambda line: f'>>>{line}<<<', None):
                self.logger.error('FOO')

        self.run_in_context(in_context)
        if self.run_in_context is run_in_no_context:
            assert self.handler.messages == ['>>>[FOO]<<<']
        else:
            assert self.handler.messages == ['[FOO]']

    def test_context_decorator_decorates_in_context(self):
        def in_context():
            with logcontext.Decorator.use(lambda line: f'>>>{line}<<<'):
                self.logger.error('FOO')

        self.run_in_context(in_context)
        assert self.handler.messages == ['>>>[FOO]<<<']

    def test_use_None_as_context_decorator(self):
        def in_context():
            with logcontext.Decorator.use(None):
                self.logger.error('FOO')

        self.run_in_context(in_context)
        assert self.handler.messages == ['[FOO]']

    def test_use_string_as_context_decorator(self):
        def in_context():
            with logcontext.Decorator.use('❰❰❰{}❱❱❱'):
                self.logger.error('FOO')

        self.run_in_context(in_context)
        assert self.handler.messages == ['❰❰❰[FOO]❱❱❱']

    def test_use_bytes_as_context_decorator(self):
        def in_context():
            with logcontext.Decorator.use(b'>>>'):
                self.logger.error('FOO')

        self.run_in_context(in_context)
        assert self.handler.messages == ['>>>[FOO]']



import asyncio
import contextlib
import io
import logging
import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory

logger = logging.getLogger(__name__)


def default_decorator(line):
    return line


def simple_decorator(pattern=None):
    """Build a simple decorator function from the given pattern.

    The decorator will prefix each line with the given pattern. If the given
    pattern contains a '{}' placeholder, it will be used as a separator
    between the prefix and suffix. In other words, this:

      simple_decorator('foo {} bar')

    is equivalent to this

      lambda bs: b'foo ' + bs + b' bar' + b'\n'

    Since StreamMux works on bytes only, the decorator function returned from
    here will also work on bytes only. However, for your convenience (and as
    shown above), any strings given to this factory will be automatically
    converted to bytes using string.encode('utf-8', errors='surrogateescape').
    """
    # process pattern as string, so we can use .format() below
    if isinstance(pattern, (bytes, bytearray)):
        pattern = pattern.decode('utf-8', errors='surrogateescape')
    elif pattern is None:
        pattern = ''
    assert isinstance(pattern, str)

    try:
        assert '\0' not in pattern
        prefix, suffix = pattern.format('\0').split('\0')
    except (AssertionError, ValueError, IndexError):
        prefix, suffix = pattern, ''

    # convert everything to bytes
    prefix = prefix.encode('utf-8', errors='surrogateescape')
    suffix = suffix.encode('utf-8', errors='surrogateescape')
    return lambda line: prefix + line.rstrip() + suffix.rstrip() + b'\n'


class DecoratedStream:
    """Manage the opening/closing of a StreamMux FIFO."""

    def __init__(self, mux, decorator=None, mode='w'):
        self.mux = mux
        self.decorator = decorator
        self.mode = mode
        self.path = None
        self.fifo = None
        self.used_by = 0

    def open(self):
        if self.path is None:  # FIFO not yet open
            assert self.fifo is None
            self.path = self.mux.watched_fifo(self.decorator)
            self.fifo = self.path.open(self.mode)
        self.used_by += 1
        return self.fifo

    def close(self):
        if self.path is None:  # already closed
            assert self.used_by == 0 and self.fifo is None
            return

        self.used_by -= 1
        if self.used_by <= 0:
            if self.fifo is not None:
                self.fifo.close()
                self.fifo = None
            self.mux.unwatch(self.path)
            self.path = None

    async def cat_file(self, path, tail=None):
        decorator = self.mux._prepare_decorator(self.decorator)
        lines = path.read_bytes().splitlines(keepends=True)
        if tail is not None:
            lines = lines[-tail:]
        for line in lines:
            self.mux.out.write(decorator(line))

    def __enter__(self):
        return self.open()

    def __exit__(self, *_):
        self.close()


class StreamMux:
    """Multiplex several output streams (w/decoration) into a single stream."""

    default_decorator = staticmethod(default_decorator)
    simple_decorator = staticmethod(simple_decorator)

    def __init__(self, out=None, tmp_base=None):
        if out is None:
            self.out = sys.stdout.buffer
        elif not isinstance(out, (io.RawIOBase, io.BufferedIOBase)):
            try:
                self.out = out.buffer
            except AttributeError:
                raise ValueError(f'Cannot find binary stream in {out}')
        else:
            self.out = out
        self.tempdir = TemporaryDirectory(dir=tmp_base, prefix='StreamMux_')
        self.fifonum = 0
        self.watches = {}  # map path -> (f, decorator, buffer)
        self.polls = {}  # map path -> (timer, path, f, decorator, buffer)

    def new_stream(self, decorator=None, mode='w'):
        return DecoratedStream(self, decorator, mode)

    def watched_fifo(self, decorator=None):
        """Create a FIFO (aka. named pipe) that is watched by StreamMux.

        Creates a FIFO in self.tempdir, and returns its path.
        The FIFO is watched by this StreamMux, so anything written into it will
        appear on this StreamMux' output (after passing through 'decorator').

        The FIFO (and the rest of self.tempdir) will be automatically removed
        on shutdown.
        """
        self.fifonum += 1
        fifopath = Path(self.tempdir.name, f'fifo{self.fifonum}')
        assert not fifopath.exists()
        os.mkfifo(fifopath)
        self._watch(fifopath, decorator)
        return fifopath

    def _prepare_decorator(self, decorator=None):
        if decorator is None:  # => identity decorator
            return self.default_decorator
        elif isinstance(decorator, (str, bytes, bytearray)):  # => pattern
            return self.simple_decorator(decorator)
        else:
            assert callable(decorator)
            return decorator

    def _watch(self, path, decorator=None):
        """Add the given 'path' to be watched by StreamMux.

        Lines read from 'path' will be passed through 'decorator' before being
        written to this StreamMux' shared output.
        """
        decorator = self._prepare_decorator(decorator)
        logger.debug(f'Watching {path}')
        assert path not in self.watches
        fd = os.open(path, os.O_RDONLY | os.O_NONBLOCK)
        f = open(fd, mode='rb', buffering=0)
        buffer = bytearray()
        self.watches[path] = f, decorator, buffer
        asyncio.get_running_loop().add_reader(
            f, self._do_read, f, decorator, buffer
        )

    def _process_lines(self, decorator, buffer, last):
        lines = buffer.splitlines(keepends=True)
        if lines and not lines[-1].endswith(b'\n') and not last:
            # keep partial line in buffer
            del buffer[: -len(lines[-1])]
            lines.pop()
        else:
            buffer.clear()
        for line in lines:
            self.out.write(decorator(line))

    def _do_read(self, f, decorator, buffer, *, last=False):
        buffer.extend(f.read() or b'')  # .read() may return None w/O_NONBLOCK
        self._process_lines(decorator, buffer, last)

    def unwatch(self, path):
        """Stop watching the given 'path'."""
        logger.debug(f'Unwatching {path}')
        assert path in self.watches
        f, decorator, buffer = self.watches.pop(path)
        asyncio.get_running_loop().remove_reader(f)
        self._do_read(f, decorator, buffer, last=True)
        f.close()
        assert not buffer

    @contextlib.contextmanager
    def follow_file(self, path, decorator=None, *, period=0.1):
        """Context manager to poll the given path while inside this context."""
        self.poll(path, decorator, period=period)
        try:
            yield
        finally:
            self.unpoll(path)

    def poll(self, path, decorator=None, *, period=0.1):
        """Add the given 'path' to be polled periodically by StreamMux.

        Lines read from 'path' will be passed through 'decorator' before being
        written to this StreamMux' shared output.
        """
        assert period > 0
        decorator = self._prepare_decorator(decorator)
        logger.debug(f'Polling {path}')
        assert path not in self.polls
        f = path.open(mode='rb')
        buffer = bytearray()
        timer = asyncio.get_running_loop().call_later(
            0, self._do_poll, path, f, decorator, buffer, period
        )
        self.polls[path] = (timer, path, f, decorator, buffer)

    def _do_poll(self, path, f, decorator, buffer, period):
        last = period < 0
        pos = f.tell()
        stat = os.stat(f.fileno())
        end = stat.st_size

        if pos < end:  # There is something to be read
            # Try to read cheaply, unless this is the last call
            chunk = f.read() if last else f.read1(stat.st_blksize)
            buffer.extend(chunk)
            pos += len(chunk)
            self._process_lines(decorator, buffer, last)

        if not last:  # Schedule next poll
            # Read again ASAP if we already know there is more to read
            delay = period if pos >= end else 0
            timer = asyncio.get_running_loop().call_later(
                delay, self._do_poll, path, f, decorator, buffer, period
            )
            self.polls[path] = (timer, path, f, decorator, buffer)

    def unpoll(self, path):
        """Stop polling the given 'path'."""
        logger.debug(f'Unpolling {path}')
        assert path in self.polls
        timer, path, f, decorator, buffer = self.polls.pop(path)
        timer.cancel()
        self._do_poll(path, f, decorator, buffer, -1)  # last call
        f.close()
        assert not buffer

    def shutdown(self):
        """Shutdown this StreamMux instance.

        Stop watching all files and cleanup temporary FIFOs.
        """
        logger.debug('Shutting down')
        for path in sorted(self.watches.keys()):
            logger.warning(f'{path} was not unwatched!')
            self.unwatch(path)
        for path in sorted(self.polls.keys()):
            logger.warning(f'{path} was not unpolled!')
            self.unpoll(path)
        if not self.out.closed:
            self.out.flush()
        self.tempdir.cleanup()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.shutdown()



import asyncio
import contextlib
import functools
import logging
import pytest
import random
import resource
from subprocess import PIPE
import sys
import time

from asyncjobs import logcontext, decorated_work, signal_handling, stream_mux

from conftest import (
    abort_in,
    assert_elapsed_time,
    ListHandler,
    mock_argv,
    TExternalWorkJob,
    verified_events,
    verify_number_of_tasks,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

logger = logging.getLogger(__name__)
testlogger = logging.getLogger('test')


def shuffled_prints(out_f, err_f, out_strings, err_strings):
    """Print out_strings -> out_f and err_strings -> err_f in shuffled order.

    Return a stream of callables that together prints out_strings -> out_f and
    err_strings -> err_f such that each is processed in order (out_strings is
    printed in order to out_f, and err_strings is printed in order to err_f),
    but the out_f and err_f prints are interleaved in a random order. E.g.
    this call:

        shuffled_prints(sys.stdout, sys.stderr, list('foo'), list('bar'))

    might result in the following (un-called) prints being yielded:

        print('b', file=sys.stderr, end='')
        print('f', file=sys.stdout, end='')
        print('o', file=sys.stdout, end='')
        print('a', file=sys.stderr, end='')
        print('o', file=sys.stdout, end='')
        print('r', file=sys.stderr, end='')

    This helper is an attempt to provoke any issues we might have regarding
    rescheduling/ordering of tasks and internal handling of file descriptors.
    """

    def prints(f, items):
        """Yield callables that print each item of items to f, in order."""
        for item in items:
            yield functools.partial(print, item, file=f, end='')

    def random_interleave(*iterators):
        """Yield from each iterator in random order until all are exhausted."""
        iterators = list(iterators)
        while iterators:
            i = random.randrange(len(iterators))
            try:
                yield next(iterators[i])
            except StopIteration:  # iterators[i] is exhausted
                del iterators[i]

    out_prints = prints(out_f, out_strings)
    err_prints = prints(err_f, err_strings)
    yield from random_interleave(out_prints, err_prints)


class TUndecoratedJob(TExternalWorkJob):
    def __init__(
        self,
        *args,
        mode=None,
        out=None,
        err=None,
        log=None,
        extras=None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        Modes = {
            'async': self.do_async,
            'thread': self.do_thread,
            'mock_argv': self.do_mock_argv,
            'custom': self.coro,
        }
        if mode is None:
            mode = 'async' if self.coro is None else 'custom'
        self.coro = Modes[mode]

        self.out = f"This is {self.name}'s stdout\n" if out is None else out
        self.err = f"This is {self.name}'s stderr\n" if err is None else err
        self.log = f"This is {self.name}'s log" if log is None else log
        self.extras = extras

    def xout(self):
        """Return expected stdout data from this job."""
        lines = self.out if isinstance(self.out, list) else [self.out]
        return [line.rstrip() for line in lines]

    def xerr(self, *, include_log=False):
        """Return expected stderr data from this job."""
        lines = self.err if isinstance(self.err, list) else [self.err]
        if include_log and self.log:
            lines += [self.log]
        return [line.rstrip() for line in lines]

    def xlog(self):
        """Return expected log messages from this job."""
        lines = self.log if isinstance(self.log, list) else [self.log]
        return [line.rstrip() for line in lines]

    def shuffled_out_err(self, out_f, err_f):
        if isinstance(self.out, list):  # lines from list of strings
            out = (line + '\n' for line in self.out)
        else:  # characters from string
            out = (char for char in self.out)
        if isinstance(self.err, list):  # lines from list of strings
            err = (line + '\n' for line in self.err)
        else:  # characters from string
            err = (char for char in self.err)
        return shuffled_prints(out_f, err_f, out, err)

    def mock_argv(self, extra_args=None):
        """Return a suitable mock_argv for doing this job in a subprocess."""
        assert self.out.endswith('\n')  # newline-terminated string
        assert self.err.endswith('\n')  # newline-terminated string
        args = ['out:', self.out.rstrip()]
        for arg in extra_args or []:
            if not arg.startswith('exit:'):
                args.append(arg)
        args += ['err:', self.err.rstrip()]
        if self.log:
            args.append(f'log:{self.log}')
        for arg in extra_args or []:
            if arg.startswith('exit:'):
                args.append(arg)
        return mock_argv(*args)

    async def do_async(self, ctx, outf=None, errf=None):
        outf = sys.stdout if outf is None else outf
        errf = sys.stderr if errf is None else errf
        for print_one in self.shuffled_out_err(outf, errf):
            print_one()
            await asyncio.sleep(0)
        if self.log:
            testlogger.error(self.log)

    async def do_thread(self, ctx, outf=None, errf=None, ticket=None):
        outf = sys.stdout if outf is None else outf
        errf = sys.stderr if errf is None else errf

        def in_thread(outf, errf, *_):
            for print_one in self.shuffled_out_err(outf, errf):
                print_one()
                time.sleep(0.001)
            if self.log:
                testlogger.error(self.log)

        return await self.run_thread(
            functools.partial(in_thread, outf, errf), ctx, ticket=ticket,
        )

    async def do_mock_argv(self, ctx):
        return await self.run_subprocess(self.mock_argv(self.extras), ctx)


class TJob(TUndecoratedJob):
    def __init__(self, *args, decorate=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.decorate = decorate

    async def __call__(self, ctx):
        with ctx.decoration(
            decorate_out=f'{self.name}/out: ' if self.decorate else None,
            decorate_err=f'{self.name}/ERR: ' if self.decorate else None,
            decorate_log=f'{self.name}/log: ' if self.decorate else None,
        ):
            return await super().__call__(ctx)

    def xout(self):
        prefix = f'{self.name}/out: ' if self.decorate else ''
        return [prefix + line for line in super().xout()]

    def xerr(self, **kwargs):
        prefix = f'{self.name}/ERR: ' if self.decorate else ''
        return [prefix + line for line in super().xerr(**kwargs)]

    def xlog(self):
        prefix = f'{self.name}/log: ' if self.decorate else ''
        return [prefix + line for line in super().xlog()]

    async def do_async(self, ctx):
        with ctx.stdout as outf, ctx.stderr as errf:
            return await super().do_async(ctx, outf, errf)

    async def do_thread(self, ctx, out_stream=None, err_stream=None):
        out_stream = ctx.stdout if out_stream is None else out_stream
        err_stream = ctx.stderr if err_stream is None else err_stream
        async with ctx.reserve_worker() as ticket:
            with out_stream as outf, err_stream as errf:
                return await super().do_thread(ctx, outf, errf, ticket)


@pytest.fixture
def Scheduler(scheduler_with_workers):
    return scheduler_with_workers(
        signal_handling.Scheduler, decorated_work.Scheduler
    )


@pytest.fixture
def run(Scheduler):
    async def _run(todo, abort_after=None, check_events=True, **kwargs):
        scheduler = Scheduler(**kwargs)
        if check_events:
            cm = verified_events(scheduler, todo)
        else:
            cm = contextlib.nullcontext()
        with cm:
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            try:
                asyncio.get_running_loop().set_debug(True)
                async with abort_in(abort_after):
                    return await scheduler.run()
            finally:
                verify_number_of_tasks(1)  # no other tasks than test itself

    return _run


# no output


async def test_no_output_from_no_jobs(run, verify_output):
    await run([])
    assert verify_output([], [], [])


async def test_no_output_from_two_jobs(run, verify_output):
    await run([TJob(name, out='', err='', log='') for name in ['foo', 'bar']])
    assert verify_output([], [], [])


# undecorated output


async def test_undecorated_charwise_output_from_two_jobs(run, verify_output):
    todo = [
        TJob(
            name,
            out=f"char by char to {name}'s stdout\n",
            err=f"char by char to {name}'s stderr\n",
            decorate=False,
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_undecorated_linewise_output_from_two_jobs(run, verify_output):
    todo = [
        TJob(
            name,
            out=['line', 'by', 'line', 'to', f"{name}'s", 'stdout'],
            err=['line', 'by', 'line', 'to', f"{name}'s", 'stderr'],
            decorate=False,
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_undecorated_follow_file_to_stdout(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        with ctx.follow_file(path):
            pass

    todo = TJob('foo', coro=coro, decorate=False)
    await run([todo])
    assert verify_output([lines], [], [])


async def test_undecorated_cat_file_to_stdout(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line', 'fourth line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        await ctx.stdout.cat_file(path)

    todo = TJob('foo', coro=coro, decorate=False)
    await run([todo])
    assert verify_output([lines], [], [])


async def test_undecorated_tail_file_to_stderr(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line', 'fourth line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        await ctx.stderr.cat_file(path, tail=2)

    todo = TJob('foo', coro=coro, decorate=False)
    await run([todo])
    assert verify_output([], [lines[-2:]], [])


# decorated output


async def test_decorated_charwise_output_from_two_jobs(run, verify_output):
    todo = [
        TJob(
            name,
            out=f"char by char to {name}'s stdout\n",
            err=f"char by char to {name}'s stderr\n",
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_decorated_linewise_output_from_two_jobs(run, verify_output):
    todo = [
        TJob(
            name,
            out=['line', 'by', 'line', 'to', f"{name}'s", 'stdout'],
            err=['line', 'by', 'line', 'to', f"{name}'s", 'stderr'],
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_decorated_output_via_custom_muxes(run, verify_output):
    todo = [TJob(name) for name in ['foo', 'bar']]
    to_stdout = stream_mux.StreamMux(sys.stdout)
    to_stderr = stream_mux.StreamMux(sys.stderr)
    await run(todo, outmux=to_stderr, errmux=to_stdout)  # flip stderr/stdout
    assert verify_output(
        [job.xerr() for job in todo],  # flipped
        [job.xout() for job in todo],  # flipped
        [job.xlog() for job in todo],
    )


async def test_decorated_output_with_custom_log_handler(run, verify_output):
    handler = ListHandler()
    handler.setFormatter(logging.Formatter('<%(message)s>'))
    testlogger.addHandler(handler)
    todo = [TJob(name) for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )
    assert sorted(handler.messages) == sorted(f'<{job.log}>' for job in todo)


async def test_decorated_output_with_context_formatter(run, verify_output):
    handler = ListHandler()
    handler.setFormatter(logcontext.Formatter('<%(message)s>'))
    testlogger.addHandler(handler)
    todo = [TJob(name) for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )
    assert sorted(handler.messages) == sorted(
        f'{job.name}/log: <{job.log}>' for job in todo
    )


async def test_decorated_output_from_spawned_jobs(run, verify_output):
    baz = TJob('baz')
    bar = TJob('bar', spawn=[baz])
    foo = TJob('foo', spawn=[bar])
    await run([foo])
    assert verify_output(
        [job.xout() for job in [foo, bar, baz]],
        [job.xerr() for job in [foo, bar, baz]],
        [job.xlog() for job in [foo, bar, baz]],
    )


async def test_decorated_follow_file_to_stdout(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        with ctx.follow_file(path):
            pass

    todo = TJob('foo', coro=coro)
    await run([todo])
    assert verify_output([[f'foo/out: {line}' for line in lines]], [], [])


async def test_decorated_follow_file_to_stderr(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        with ctx.follow_file(path, use_stderr=True):
            pass

    todo = TJob('foo', coro=coro)
    await run([todo])
    assert verify_output([], [[f'foo/ERR: {line}' for line in lines]], [])


async def test_decorated_tail_file_to_stdout(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line', 'fourth line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        await ctx.stdout.cat_file(path, tail=2)

    todo = TJob('foo', coro=coro)
    await run([todo])
    assert verify_output([[f'foo/out: {line}' for line in lines[-2:]]], [], [])


async def test_decorated_cat_file_to_stderr(tmp_path, run, verify_output):
    lines = ['first line', 'second line', 'third line', 'fourth line']
    path = tmp_path / 'file'
    path.write_text(''.join(line + '\n' for line in lines))

    async def coro(ctx):
        await ctx.stderr.cat_file(path)

    todo = TJob('foo', coro=coro)
    await run([todo])
    assert verify_output([], [[f'foo/ERR: {line}' for line in lines]], [])


# decorated output from threads


async def test_decorated_output_from_thread(run, verify_output):
    todo = [TJob(name, mode='thread') for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_decorated_output_from_thread_via_follow_file(
    tmp_path, run, verify_output
):
    # Use TJob.do_thread(), but redirect to files instead of FIFOs
    async def coro(ctx):
        outpath = tmp_path / f'{ctx.tjob.name}_outfile'
        errpath = tmp_path / f'{ctx.tjob.name}_errfile'
        with outpath.open('w') as outf, errpath.open('w') as errf:
            with ctx.follow_file(outpath), ctx.follow_file(errpath, True):
                return await ctx.tjob.do_thread(ctx, outf, errf)

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


# decorated output from subprocesses


async def test_decorated_output_from_subprocess_worker(run, verify_output):
    todo = [TJob(name, mode='mock_argv') for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr(include_log=True) for job in todo],
    )


async def test_decorated_output_from_subprocess_via_follow_file(
    tmp_path, run, verify_output
):
    async def coro(ctx):
        outpath = tmp_path / f'{ctx.tjob.name}_outfile'
        errpath = tmp_path / f'{ctx.tjob.name}_errfile'
        argv = ctx.tjob.mock_argv()
        with outpath.open('w') as outf, errpath.open('w') as errf:
            with ctx.follow_file(outpath), ctx.follow_file(errpath, True):
                with ctx.tjob.subprocess_xevents(argv):
                    return await ctx.run_in_subprocess(
                        argv, stdout=outf, stderr=errf, check=True
                    )

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr(include_log=True) for job in todo],
    )


async def test_decorated_output_from_aborted_processes(
    num_workers, run, verify_output
):
    todo = [
        TJob(name, mode='mock_argv', extras=['sleep:5'])
        for name in ['foo', 'bar', 'baz']
    ]
    with assert_elapsed_time(lambda t: t < 0.75):
        await run(todo, abort_after=0.5)

    # We have 3 jobs, but can only run as many concurrently as there are
    # workers available. The rest will be cancelled before they start.
    assert verify_output(
        [job.xout() for job in todo][:num_workers],
        [],  # No stderr output as this happens _after_ the aborted sleep
    )


async def test_decorated_output_from_subprocess_context(run, verify_output):
    async def coro(ctx):
        argv = ctx.tjob.mock_argv()
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv) as proc:
                await proc.wait()

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr(include_log=True) for job in todo],
    )


async def test_subproc_decorate_stderr_capture_stdout(run, verify_output):
    async def coro(ctx):
        argv = ctx.tjob.mock_argv()
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                output = await proc.stdout.read()
                await proc.wait()
                return output

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    done = await run(todo)
    assert verify_tasks(  # undecorated output was captured and returned
        done, {job.name: job.out.encode('ascii') for job in todo}
    )
    assert verify_output(
        [], [job.xerr(include_log=True) for job in todo],  # output captured
    )


async def test_subproc_decorate_stderr_via_follow_file_capture_stdout(
    tmp_path, run, verify_output
):
    async def coro(ctx):
        outpath = tmp_path / f'{ctx.tjob.name}_outfile'
        errpath = tmp_path / f'{ctx.tjob.name}_errfile'
        argv = ctx.tjob.mock_argv()
        with outpath.open('w'), errpath.open('w') as errf:
            with ctx.follow_file(outpath), ctx.follow_file(errpath, True):
                with ctx.tjob.subprocess_xevents(argv, result=0):
                    async with ctx.subprocess(
                        argv, stdout=PIPE, stderr=errf
                    ) as proc:
                        output = await proc.stdout.read()
                        await proc.wait()
                        return output

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    done = await run(todo)
    assert verify_tasks(  # undecorated output was captured and returned
        done, {job.name: job.out.encode('ascii') for job in todo}
    )
    assert verify_output(
        [], [job.xerr(include_log=True) for job in todo],  # output captured
    )


async def test_subproc_capture_stdout_from_terminated_proc(run, verify_output):
    async def coro(ctx):
        argv = ctx.tjob.mock_argv(['sleep:30'])
        with ctx.tjob.subprocess_xevents(argv, result='terminate'):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                return await proc.stdout.readline()
                # Skipping await proc.wait() to provoke termination

    todo = [TJob(name, coro=coro) for name in ['foo', 'bar']]
    done = await run(todo)
    assert verify_tasks(  # undecorated output was captured and returned
        done, {job.name: job.out.encode('ascii') for job in todo}
    )
    assert verify_output([], [])  # subproc terminated before print to stderr


# output outside decoration context


async def test_output_outside_decoration_from_coro(run, verify_output):
    todo = [
        TUndecoratedJob(
            name,
            out=[f"This is {name}'s stdout"],  # Use lines instead of chars to
            err=[f"This is {name}'s stderr"],  # prevent jumbled stdout/stderr.
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_output_outside_decoration_from_thread(run, verify_output):
    todo = [
        TUndecoratedJob(
            name,
            mode='thread',
            out=[f"This is {name}'s stdout"],  # Use lines instead of chars to
            err=[f"This is {name}'s stderr"],  # prevent jumbled stdout/stderr.
        )
        for name in ['foo', 'bar']
    ]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_output_outside_decoration_from_subprocess(run, verify_output):
    todo = [TUndecoratedJob(name, mode='mock_argv') for name in ['foo', 'bar']]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr(include_log=True) for job in todo],
    )


async def test_follow_file_outside_decoration_raises_RuntimeError(
    tmp_path, run, verify_output
):
    path = tmp_path / 'file'
    path.write_text('\n'.join(['first line', 'second line', 'third line']))

    async def coro(ctx):
        with ctx.follow_file(path):
            pass

    todo = TUndecoratedJob('foo', coro=coro)
    done = await run([todo])
    assert verify_tasks(
        done,
        {'foo': RuntimeError('Cannot follow file outside decoration context')},
    )


# decorated_job() decorator


async def test_decorated_job_with_no_decoration(run, verify_output):
    @decorated_work.decorated_job()
    async def job(ctx):
        with ctx.stdout as outf, ctx.stderr as errf:
            print('Printing to stdout', file=outf)
            print('Printing to stderr', file=errf)
            testlogger.error('Logging')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['Printing to stdout']], [['Printing to stderr']], [['Logging']],
    )


async def test_decorated_job_with_decoration(run, verify_output):
    @decorated_work.decorated_job('foo/out: ', 'foo/ERR: ', 'foo/log: ')
    async def job(ctx):
        with ctx.stdout as outf, ctx.stderr as errf:
            print('Printing to stdout', file=outf)
            print('Printing to stderr', file=errf)
            testlogger.error('Logging')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['foo/out: Printing to stdout']],
        [['foo/ERR: Printing to stderr']],
        [['foo/log: Logging']],
    )


async def test_decorated_job_only_decorate_stderr_and_log(run, verify_output):
    @decorated_work.decorated_job(decorate_err='foo/ERR: ', decorate_log=True)
    async def job(ctx):
        with ctx.stdout as outf, ctx.stderr as errf:
            print('Printing to stdout', file=outf)
            print('Printing to stderr', file=errf)
            testlogger.error('Logging')

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['Printing to stdout']],
        [['foo/ERR: Printing to stderr']],
        [['foo/ERR: Logging']],
    )


async def test_decorated_job_with_decorated_subprocess(run, verify_output):
    @decorated_work.decorated_job('foo/out: ', 'foo/ERR: ', 'foo/log: ')
    async def job(ctx):
        argv = mock_argv(
            'Printing to stdout', 'err:', 'Printing to stderr', 'log:Logging!'
        )
        async with ctx.subprocess(argv) as proc:
            await proc.wait()

    job.name = 'foo'
    await run([job], check_events=False)
    assert verify_output(
        [['foo/out: Printing to stdout']],
        [['foo/ERR: Printing to stderr', 'foo/ERR: Logging!']],
        [],  # log output from subprocess goes to its stderr
    )


async def test_decorated_job_with_subproc_output_capture(run, verify_output):
    @decorated_work.decorated_job('foo/out: ', 'foo/ERR: ', 'foo/log: ')
    async def job(ctx):
        argv = mock_argv(
            'Printing to stdout', 'err:', 'Printing to stderr', 'log:Logging!'
        )
        async with ctx.subprocess(argv, stdout=PIPE) as proc:
            output = await proc.stdout.read()
            await proc.wait()
            return output

    job.name = 'foo'
    done = await run([job], check_events=False)
    assert verify_tasks(done, {'foo': b'Printing to stdout\n'})  # undecorated
    assert verify_output(
        [],  # output was captured
        [['foo/ERR: Printing to stderr', 'foo/ERR: Logging!']],
        [],  # log output from subprocess goes to its stderr
    )


# stress-testing the decorated_work framework


async def test_decorated_output_from_many_jobs(num_jobs, run, verify_output):
    todo = [TJob(f'job #{i}') for i in range(num_jobs)]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_decorated_output_from_many_spawned_jobs(
    num_jobs, run, verify_output
):
    todo = TJob(
        'foo',
        out='',
        err='',
        log='',
        spawn=[TJob(f'job #{i}') for i in range(num_jobs)],
    )
    await run([todo])
    assert verify_output(
        [job.xout() for job in todo.spawn],
        [job.xerr() for job in todo.spawn],
        [job.xlog() for job in todo.spawn],
    )


async def test_decorated_output_from_many_thread_workers(
    num_jobs, run, verify_output
):
    todo = [TJob(f'job #{i}', mode='thread') for i in range(num_jobs)]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_decorated_output_from_many_subprocesses(
    num_jobs, run, verify_output
):
    todo = [TJob(f'job #{i}', mode='mock_argv') for i in range(num_jobs)]
    await run(todo)
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr(include_log=True) for job in todo],
    )


# Attempt to exhaust available file descriptors

# Each job may open 2 FIFOs, which both have a reader and a writer, so with
# everything opened, we use ~4 fds per job (we tend to run out at ~252 jobs
# on systems with 1024 fds available). Limit number of jobs to _half_ the
# number of available fds to reliably trigger test failures when we open
# too many fds concurrently.
fd_limit = resource.getrlimit(resource.RLIMIT_NOFILE)[0]


async def test_provoke_too_many_open_files(num_workers, run):
    # We will run out of file desciptors if we indiscriminately open
    # stdout/stderr for every single job, without any throttling.
    if num_workers < 10:
        pytest.skip('not worth running on Schedulers with few workers')

    def task_cancelled_or_failed_with_EMFILE(task):
        if task.cancelled():
            return True
        e = task.exception()
        if not isinstance(e, OSError):
            return False
        return e.args == (24, 'Too many open files')

    todo = [TJob(f'job #{i}') for i in range(fd_limit // 4)]
    done = await run(todo)
    assert verify_tasks(
        done, {job.name: task_cancelled_or_failed_with_EMFILE for job in todo}
    )


async def test_logging_only_does_not_exhaust_open_files(
    num_workers, run, verify_output
):
    if num_workers < 10:
        pytest.skip('not worth running on Schedulers with few workers')

    async def coro(ctx):
        testlogger.error(ctx.tjob.log)

    todo = [TJob(f'job #{i}', coro=coro) for i in range(fd_limit // 4)]
    done = await run(todo)
    assert all(t.done() and t.result() is None for t in done.values())
    assert verify_output([], [], [job.xlog() for job in todo])  # no stdout/err


async def test_use_reserve_worker_to_limit_open_fifos(
    num_workers, run, verify_output
):
    if num_workers < 10:
        pytest.skip('not worth running on Schedulers with few workers')

    async def coro(ctx):
        # Use ctx.reserve_worker() to prevent fd exhaustion
        async with ctx.reserve_worker():
            return await ctx.tjob.do_async(ctx)

    todo = [TJob(f'job #{i}', coro=coro) for i in range(fd_limit // 4)]
    done = await run(todo, check_events=False)
    assert all(t.done() and t.result() is None for t in done.values())
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_use_threads_to_limit_open_fifos(
    num_workers, run, verify_output
):
    if num_workers < 10:
        pytest.skip('not worth running on Schedulers with few workers')
    if fd_limit > 1024:
        pytest.skip(f'skip trying to exhaust {fd_limit} file descriptors')

    todo = [TJob(f'job #{i}', mode='thread') for i in range(fd_limit // 4)]
    done = await run(todo, check_events=False)
    assert all(t.done() and t.result() is None for t in done.values())
    assert verify_output(
        [job.xout() for job in todo],
        [job.xerr() for job in todo],
        [job.xlog() for job in todo],
    )


async def test_use_subprocesses_to_limit_open_fifos(
    num_workers, run, verify_output
):
    if num_workers < 10:
        pytest.skip('not worth running on Schedulers with few workers')
    if fd_limit > 1024:
        pytest.skip(f'skip trying to exhaust {fd_limit} file descriptors')

    async def coro(ctx):
        return await ctx.run_in_subprocess(['true'])

    todo = [TJob(f'job #{i}', coro=coro) for i in range(fd_limit // 4)]
    done = await run(todo, check_events=False)
    assert all(t.done() and t.result() == 0 for t in done.values())
    assert verify_output([], [])  # no output from `true`



import asyncio
import concurrent.futures
import contextlib
import logging
import subprocess

from . import basic

logger = logging.getLogger(__name__)


class Context(basic.Context):
    """Extend Context with helpers for doing work in threads + processes."""

    _next_worker_ticket = 0

    @classmethod
    def next_worker_ticket(cls):
        ret = cls._next_worker_ticket
        cls._next_worker_ticket += 1
        return ret

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._workers_reserved = set()
        self._workers_in_use = set()

    @contextlib.asynccontextmanager
    async def reserve_worker(self):
        """Acquire a worker context where this job can run its own work.

        This is the mechanism that ensures we do not schedule more concurrent
        work than allowed by Scheduler.workers. Anybody that wants to spin off
        another thread or process to perform some work should use this context
        manager to await a free "slot".

        Yields a worker "ticket" into the context, which can be passed to other
        methods that would otherwise need to reserve their own worker ticket.

        If a single job tries to reserve more concurrent workers than allowed
        by Scheduler.workers, a RuntimeError will be raised.
        """
        max_reserve = self._scheduler.workers
        if len(self._workers_reserved) >= max_reserve:
            raise RuntimeError(f'Cannot reserve >={max_reserve} worker(s)!')
        logger.debug(f'{self.name} -> acquiring worker ticket…')
        self.event('await worker slot')
        async with self._scheduler.worker_semaphore:
            ticket = self.next_worker_ticket()
            self._workers_reserved.add(ticket)
            self.event('awaited worker slot')
            logger.debug(f'{self.name} -- acquired worker ticket {ticket}')
            try:
                yield ticket
            finally:
                self._workers_reserved.remove(ticket)
                logger.debug(f'{self.name} <- released worker ticket {ticket}')

    @contextlib.asynccontextmanager
    async def _use_ticket(self, ticket=None, only_reserve=False):
        """Reuse the given worker ticket or reserve a new one."""
        if ticket is None:  # issue new ticket
            async with self.reserve_worker() as ticket:
                yield ticket
        else:  # verify ticket
            if ticket not in self._workers_reserved:
                raise ValueError(f'Cannot use ticket {ticket}: Not reserved')
            elif ticket in self._workers_in_use:
                raise ValueError(f'Cannot use ticket {ticket}: Already in use')
            try:
                if not only_reserve:
                    self._workers_in_use.add(ticket)
                yield ticket
            finally:
                if not only_reserve:
                    self._workers_in_use.remove(ticket)

    async def call_in_thread(self, func, *args, ticket=None):
        """Call func(*args) in a worker thread and await its result.

        Reuse worker ticket if given, otherwise reserve a new ticket.
        """
        async with self._use_ticket(ticket):
            try:
                logger.debug(f'{self.name} -> start {func} in worker thread…')
                self.event('start work in thread', func=str(func))
                future = self._scheduler._start_in_thread(func, *args)
                future.add_done_callback(
                    lambda fut: self.event(
                        'finish work in thread',
                        fate=self._scheduler._fate(fut),
                    )
                )
                logger.debug(f'{self.name} -- awaiting worker thread…')
                result = await future
                logger.debug(f'{self.name} <- {result!r} from worker')
                return result
            except Exception as e:
                logger.warning(f'{self.name} <- Exception {e!r} from worker!')
                raise

    async def terminate_subprocess(self, proc, argv, delay, *, kill=False):
        if proc.returncode is not None:  # process has already exited
            return

        verb = 'kill' if kill else 'terminate'
        logger.warning(f'{self.name}: {proc} is still alive, {verb}…')
        self.event(f'subprocess {verb}', argv=argv, pid=proc.pid)
        with contextlib.suppress(ProcessLookupError):
            if kill:
                proc.kill()
            else:
                proc.terminate()
        try:
            out, err = await asyncio.wait_for(proc.communicate(), delay)
            if out:
                logger.warning(f'{self.name}: {proc} stdout discard: {out!r}')
            if err:
                logger.warning(f'{self.name}: {proc} stderr discard: {err!r}')
            logger.debug(f'{self.name}: {proc} {verb} done.')
        except asyncio.TimeoutError:
            logger.warning(f'{self.name}: {delay}s timeout on {verb} {proc}')
        finally:
            if proc.returncode is None:  # still running
                await self.terminate_subprocess(proc, argv, delay, kill=True)

    @contextlib.asynccontextmanager
    async def subprocess(
        self, argv, *, check=False, kill_delay=3, ticket=None, **kwargs
    ):
        """Run a command line in a subprocess and interact with it.

        This context manager yields the asyncio.subprocess.Process instance
        into the context, and allows the context to interact with the process.
        The context should await the completion of the subprocess, as the
        process (if still alive) will be automatically killed upon exiting the
        context.

        Killing the subprocess is done by sending SIGTERM to it, and if it's
        still alive 3 seconds later, send SIGKILL as well. The delay between
        these signals can be customized with the 'kill_delay' argument.

        If check=True (the default is False), and the subprocess exits with a
        non-zero exit code (this includes the case where we have to kill it),
        we will raise subprocess.CalledProcessError. Otherwise, a non-zero exit
        code is not considered exceptional.

        Reuse worker ticket if given, otherwise reserve a new ticket.
        """
        async with self._use_ticket(ticket):
            logger.debug(f'{self.name} -> start {argv} in subprocess…')
            self.event('start work in subprocess', argv=argv)
            proc = await asyncio.create_subprocess_exec(*argv, **kwargs)
            try:
                logger.debug(f'{self.name} -- enter subprocess context…')
                yield proc
            finally:
                logger.debug(f'{self.name} -- exited subprocess context…')
                try:
                    if proc.returncode is None:  # still running
                        await self.terminate_subprocess(proc, argv, kill_delay)
                finally:
                    assert proc.returncode is not None
                    self.event(
                        'finish work in subprocess', returncode=proc.returncode
                    )
            if check and proc.returncode != 0:
                raise subprocess.CalledProcessError(proc.returncode, argv)

    async def run_in_subprocess(self, argv, **kwargs):
        """Run a command line in a subprocess and await its exit code."""
        async with self.subprocess(argv, **kwargs) as proc:
            logger.debug(f'{self.name} -- awaiting subprocess…')
            await proc.wait()
            logger.debug(f'{self.name} -- awaited subprocess')
            return proc.returncode


class Scheduler(basic.Scheduler):
    """Manage jobs whose work is done in threads or subprocesses.

    Extend Scheduler with methods that allow job coroutines to perform work in
    other threads or subprocesses, while keeping the number of _concurrent_
    threads/processes within the given limit.
    """

    def __init__(self, *, workers=1, context_class=Context, **kwargs):
        assert workers > 0
        self.workers = workers
        self.worker_semaphore = None
        self.worker_threads = None

        assert issubclass(context_class, Context)
        super().__init__(context_class=context_class, **kwargs)

    def add_thread_job(self, name, func, *args, deps=None, **kwargs):
        async def coro(ctx):
            return await ctx.call_in_thread(func, *args, **kwargs)

        return self.add_job(name, coro, deps)

    def add_subprocess_job(self, name, argv, *, deps=None, **kwargs):
        async def coro(ctx):
            return await ctx.run_in_subprocess(argv, **kwargs)

        return self.add_job(name, coro, deps)

    def _start_in_thread(self, func, *args):
        if self.worker_threads is None:
            self.worker_threads = concurrent.futures.ThreadPoolExecutor(
                max_workers=self.workers, thread_name_prefix='Worker Thread',
            )
        return asyncio.get_running_loop().run_in_executor(
            self.worker_threads, func, *args
        )

    async def _run_tasks(self, *args, **kwargs):
        self.worker_semaphore = asyncio.BoundedSemaphore(self.workers)

        try:
            await super()._run_tasks(*args, **kwargs)
            if self.worker_threads is not None:
                logger.debug('Shutting down worker threads…')
                self.worker_threads.shutdown()  # wait=timeout is None)
                self.worker_threads = None
                logger.debug('Shut down worker threads')
        finally:
            if self.worker_threads is not None:
                logger.error('Cancelling without awaiting workers…')
                self.worker_threads.shutdown(wait=False)
                self.worker_threads = None



import pytest

from asyncjobs import basic

from conftest import (
    Cancelled,
    TBasicJob,
    verified_events,
    verify_number_of_tasks,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

TJob = TBasicJob


@pytest.fixture
def Scheduler():
    return basic.Scheduler


@pytest.fixture
def run(Scheduler):
    async def _run(todo, **run_args):
        scheduler = Scheduler()
        with verified_events(scheduler, todo):
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            try:
                return await scheduler.run(**run_args)
            finally:
                verify_number_of_tasks(1)  # no other tasks than test itself

    return _run


# simple scenarios, no dependencies


async def test_zero_jobs_does_nothing(run):
    assert await run([]) == {}


async def test_one_ok_job(run):
    todo = [TJob('foo')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done'})


async def test_one_failed_job(run):
    done = await run([TJob('foo', result=ValueError('UGH'))])
    assert len(done) == 1
    e = done['foo'].exception()
    assert isinstance(e, ValueError) and e.args == ('UGH',)
    with pytest.raises(ValueError, match='UGH'):
        done['foo'].result()


async def test_cannot_add_second_job_with_same_name(run):
    with pytest.raises(ValueError):
        await run([TJob('foo'), TJob('foo')])


async def test_job_with_nonexisting_dependency_raises_KeyError(run):
    done = await run([TJob('foo', {'MISSING'})])
    with pytest.raises(KeyError, match='MISSING'):
        done['foo'].result()


async def test_two_independent_ok_jobs(run):
    todo = [TJob('foo'), TJob('bar')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


# jobs with dependencies


async def test_one_ok_before_another_ok_job(run):
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


async def test_one_ok_before_one_failed_job(run):
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}, result=ValueError('UGH')),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': ValueError('UGH')})


async def test_one_failed_job_before_one_ok_job_cancels_second_job(run):
    todo = [
        TJob('foo', before={'bar'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': Cancelled})


async def test_one_failed_job_before_two_ok_jobs_cancels_two_jobs(run):
    todo = [
        TJob('foo', before={'bar', 'baz'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}),
        TJob('baz', {'foo'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': ValueError('UGH'), 'bar': Cancelled, 'baz': Cancelled}
    )


async def test_one_failed_job_before_two_dependent_jobs_cancels_two_jobs(run):
    todo = [
        TJob('foo', before={'bar'}, result=ValueError('UGH')),
        TJob('bar', {'foo'}, before={'baz'}),
        TJob('baz', {'bar'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': ValueError('UGH'), 'bar': Cancelled, 'baz': Cancelled}
    )


async def test_one_failed_job_between_two_ok_jobs_cancels_last_job(run):
    todo = [
        TJob('foo', before={'bar'}),
        TJob('bar', {'foo'}, before={'baz'}, result=ValueError('UGH')),
        TJob('baz', {'bar'}),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo done', 'bar': ValueError('UGH'), 'baz': Cancelled}
    )


# keep_going=False cancels jobs even when no dependency


async def test_one_ok_one_failed_job_without_keep_going_cancels_ok_job(run):
    todo = [
        TJob('foo', result=ValueError('UGH')),
        TJob('bar', async_sleep=0.01),  # allow time for potential cancellation
    ]
    done = await run(todo, keep_going=False)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': Cancelled})


async def test_one_ok_and_one_failed_job_with_keep_going_runs_ok_job(run):
    todo = [
        TJob('foo', result=ValueError('UGH')),
        TJob('bar', async_sleep=0.01),  # allow time for potential cancellation
    ]
    done = await run(todo, keep_going=True)
    assert verify_tasks(done, {'foo': ValueError('UGH'), 'bar': 'bar done'})


# jobs that spawn child jobs


async def test_one_job_spawns_another(run):
    todo = [TJob('foo', spawn=[TJob('bar')])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})


async def test_one_job_spawns_two_with_deps(run):
    # foo start bar and baz, baz depends on bar, foo waits for both to finish
    todo = [
        TJob(
            'foo',
            spawn=[
                TJob('bar', before={'foo', 'baz'}),
                TJob('baz', {'bar'}, before={'foo'}),
            ],
            await_spawn=True,
        )
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo done', 'bar': 'bar done', 'baz': 'baz done'}
    )


async def test_one_job_spawns_failing_job(run):
    todo = [TJob('foo', spawn=[TJob('bar', result=ValueError('UGH'))])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': ValueError('UGH')})


async def test_job_is_cancelled_when_waiting_for_failing_spawn(run):
    todo = [
        TJob(
            'foo',
            spawn=[TJob('bar', result=ValueError('UGH'))],
            await_spawn=True,
        )
    ]
    done = await run(todo)
    assert verify_tasks(done, {'foo': Cancelled, 'bar': ValueError('UGH')})


async def test_spawn_outliving_parent_is_not_cancelled_by_scheduler(run):
    todo = [TJob('foo', spawn=[TJob('bar', async_sleep=0.01)])]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo done', 'bar': 'bar done'})



import asyncio
import contextlib
import logging
import threading

NotSet = object()


def current_context():
    """Return a hashable identifier for the current log context.

    Use the id of the current task in async contexts, or the id of the
    current thread in threaded contexts. Otherwise return None.
    """
    obj = None
    if threading.current_thread() is not threading.main_thread():
        # we're in a (worker) thread
        obj = threading.current_thread()
    else:  # log context is determined by current task
        with contextlib.suppress(RuntimeError):
            obj = asyncio.current_task()
    return None if obj is None else id(obj)


class Decorator:
    registry = {}  # context -> decorator function

    @classmethod
    def from_string(cls, pattern):
        try:
            assert '\0' not in pattern
            prefix, suffix = pattern.format('\0').split('\0')
        except (AssertionError, ValueError, IndexError):
            prefix, suffix = pattern, ''

        return lambda line: prefix + line.rstrip() + suffix.rstrip()

    @classmethod
    def from_bytes(cls, pattern):
        return cls.from_string(
            pattern.decode('utf-8', errors='surrogateescape')
        )

    @classmethod
    def add(cls, decorator, context=NotSet):
        context = current_context() if context is NotSet else context
        if decorator is None:  # no decoration
            return cls.remove(context)
        elif isinstance(decorator, str):  # 'prefix{}suffix'
            decorator = cls.from_string(decorator)
        elif isinstance(decorator, (bytes, bytearray)):  # b'prefix{}suffix'
            decorator = cls.from_bytes(decorator)
        else:
            assert callable(decorator)
        cls.registry[context] = decorator

    @classmethod
    def get(cls, context=NotSet, *, default=None):
        context = current_context() if context is NotSet else context
        return cls.registry.get(context, default)

    @classmethod
    def remove(cls, context=NotSet):
        context = current_context() if context is NotSet else context
        return cls.registry.pop(context, None)

    @classmethod
    @contextlib.contextmanager
    def use(cls, decorator, context=NotSet):
        context = current_context() if context is NotSet else context
        cls.add(decorator, context)
        try:
            yield
        finally:
            cls.remove(context)

    @classmethod
    def apply(cls, line, context=NotSet):
        return cls.get(context, default=lambda line: line)(line)


class Formatter(logging.Formatter):
    """Decorate log records with the registered context decorators."""

    def format(self, record):
        context = getattr(record, 'context', current_context())
        return Decorator.apply(super().format(record), context)



import asyncio
import os
from pathlib import Path
import pytest
import re
from subprocess import CalledProcessError, PIPE, STDOUT
import time

from asyncjobs import external_work

from conftest import (
    assert_elapsed_time,
    Cancelled,
    mock_argv,
    TExternalWorkJob,
    verified_events,
    verify_number_of_tasks,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

TJob = TExternalWorkJob


@pytest.fixture
def Scheduler(scheduler_with_workers):
    return scheduler_with_workers(external_work.Scheduler)


@pytest.fixture
def run(Scheduler):
    async def _run(todo):
        scheduler = Scheduler()
        with verified_events(scheduler, todo):
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            try:
                return await scheduler.run()
            finally:
                verify_number_of_tasks(1)  # no other tasks than test itself

    return _run


async def copy_lines(src, dst):
    while True:
        line = await src.readline()
        if not line:
            dst.close()
            break
        dst.write(line)
        await dst.drain()


# simple thread scenarios with call_in_thread() helper


async def test_one_ok_job_in_thread(run):
    todo = [TJob('foo', thread=lambda ctx: 'foo worked')]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 'foo worked'})


async def test_one_failed_between_two_ok_jobs_in_threads_cancels_last(run):
    def raiseUGH(ctx):
        raise ValueError('UGH')

    todo = [
        TJob('foo', before={'bar'}, thread=lambda ctx: 'foo worked'),
        TJob('bar', {'foo'}, before={'baz'}, thread=raiseUGH),
        TJob('baz', {'bar'}, thread=lambda ctx: 'baz worked'),
    ]
    done = await run(todo)
    assert verify_tasks(
        done, {'foo': 'foo worked', 'bar': ValueError('UGH'), 'baz': Cancelled}
    )


# simple subprocess scenarios with run_in_subprocess() helper


async def test_one_ok_job_in_subproc(run, tmp_path):
    path = tmp_path / 'foo'
    todo = [TJob('foo', argv=mock_argv(f'touch:{path}'))]
    done = await run(todo)
    assert verify_tasks(done, {'foo': 0})
    assert path.is_file()


async def test_one_failed_between_two_in_subprocs_cancels_last(run, tmp_path):
    foo_path = tmp_path / 'foo'
    baz_path = tmp_path / 'baz'
    todo = [
        TJob('foo', before={'bar'}, argv=mock_argv(f'touch:{foo_path}')),
        TJob('bar', {'foo'}, before={'baz'}, argv=mock_argv('exit:1')),
        TJob('baz', {'bar'}, argv=mock_argv(f'touch:{baz_path}')),
    ]
    done = await run(todo)
    assert verify_tasks(
        done,
        {
            'foo': 0,
            'bar': CalledProcessError(1, mock_argv('exit:1')),
            'baz': Cancelled,
        },
    )
    assert foo_path.is_file()
    assert not baz_path.exists()


# using the subprocess context manager for more advanced scenarios


async def test_one_ok_subprocess_with_output(run):
    argv = mock_argv('FOO')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                output = await proc.stdout.read()
                await proc.wait()
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_one_ok_subprocess_with_error_output(run):
    argv = mock_argv('err:', 'FOO')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv, stderr=PIPE) as proc:
                output = await proc.stderr.read()
                await proc.wait()
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_one_ok_subprocess_with_error_and_output_redirected(run):
    argv = mock_argv('err:', 'FOO', 'out:', 'BAR')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(
                argv, stdout=PIPE, stderr=STDOUT
            ) as proc:
                output = await proc.stdout.read()
                await proc.wait()
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\nBAR\n'})


async def test_one_failing_subprocess_with_output_default_check(run):
    argv = mock_argv('FOO', 'exit:1')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=1):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                output = await proc.stdout.read()
                await proc.wait()
                assert proc.returncode == 1
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_one_failing_subprocess_with_output_disable_check(run):
    argv = mock_argv('FOO', 'exit:1')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=1):
            async with ctx.subprocess(argv, stdout=PIPE, check=False) as proc:
                output = await proc.stdout.read()
                await proc.wait()
                assert proc.returncode == 1
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_one_failing_subprocess_with_output_enable_check(run):
    argv = mock_argv('FOO', 'exit:1')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=1):
            async with ctx.subprocess(argv, stdout=PIPE, check=True) as proc:
                output = await proc.stdout.read()
                assert output == b'FOO\n'
                await proc.wait()
                assert proc.returncode == 1

        assert False  # skipped due to CalledProcessError on context exit

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': CalledProcessError(1, argv)})


async def test_subprocess_with_custom_cwd(run, tmp_path):
    argv = mock_argv('cwd:')
    cwd = tmp_path / 'custom'
    cwd.mkdir()

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv, stdout=PIPE, cwd=cwd) as proc:
                output = await proc.stdout.read()
                await proc.wait()
        return Path(os.fsdecode(output.rstrip()))

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': cwd})


async def test_subprocess_with_custom_env(run):
    argv = mock_argv('env:FOOBAR')
    env = os.environ.copy()
    env['FOOBAR'] = 'BAZZLE'

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result=0):
            async with ctx.subprocess(argv, stdout=PIPE, env=env) as proc:
                output = await proc.stdout.read()
                await proc.wait()
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'BAZZLE\n'})


# proper termination/killing on subprocess misbehavior


async def test_not_awaiting_subprocess_terminates_it(run):
    argv = mock_argv('FOO', 'sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='terminate'):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                output = await proc.stdout.readline()
                # MISSING await proc.wait() here to trigger termination
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_partial_read_from_subprocess_before_terminating_it(run):
    argv = mock_argv('FOO', 'BAR', 'sleep:5', 'BAZ')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='terminate'):
            async with ctx.subprocess(argv, stdout=PIPE) as proc:
                output = await proc.stdout.readline()
                output += await proc.stdout.readline()
                # MISSING await proc.wait() here to trigger termination
        return output

    todo = [TJob('foo', coro=coro)]
    done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\nBAR\n'})


async def test_non_terminating_subprocess_is_killed(run):
    argv = mock_argv('ignore:SIGTERM', 'FOO', 'sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='kill'):
            async with ctx.subprocess(
                argv, stdout=PIPE, kill_delay=0.1
            ) as proc:
                output = await proc.stdout.readline()
                # MISSING await proc.wait() here to trigger termination
        return output

    todo = [TJob('foo', coro=coro)]
    with assert_elapsed_time(lambda t: t < 0.5):
        done = await run(todo)
    assert verify_tasks(done, {'foo': b'FOO\n'})


async def test_killing_subprocess_w_check_raises_CalledProcessError(run):
    argv = mock_argv('ignore:SIGTERM', 'FOO', 'sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='kill'):
            async with ctx.subprocess(
                argv, stdout=PIPE, check=True, kill_delay=0.1
            ) as proc:
                output = await proc.stdout.readline()
                # MISSING await proc.wait() here to trigger termination
        return output

    todo = [TJob('foo', coro=coro)]
    with assert_elapsed_time(lambda t: t < 0.5):
        done = await run(todo)
    assert verify_tasks(done, {'foo': CalledProcessError(-9, argv)})


# multiple threads/subprocesses from a single job


async def test_two_threads_sequentially_on_same_ticket_succeeds(run):
    def func(ret):
        time.sleep(0.001)
        return ret

    async def coro(ctx):
        async with ctx.reserve_worker() as ticket:
            with ctx.tjob.thread_xevents():
                result = await ctx.call_in_thread(func, 'FOO', ticket=ticket)
            with ctx.tjob.thread_xevents(reuse_ticket=True):
                result += await ctx.call_in_thread(func, 'BAR', ticket=ticket)
            return result

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': 'FOOBAR'})


def expect_ValueError_already_in_use(task):
    if not isinstance(task.exception(), ValueError):
        return False
    (msg,) = task.exception().args
    return re.match(r'Cannot use ticket \d+: Already in use$', msg)


async def test_two_threads_concurrently_on_same_ticket_fails(run):
    def func(ret):
        time.sleep(0.001)
        return ret

    async def coro(ctx):
        async with ctx.reserve_worker() as ticket:
            with ctx.tjob.thread_xevents():  # Only one thread started
                results = await asyncio.gather(
                    ctx.call_in_thread(func, 'FOO', ticket=ticket),
                    ctx.call_in_thread(func, 'BAR', ticket=ticket),
                    return_exceptions=True,
                )
            for result in results:
                if isinstance(result, BaseException):
                    raise result
            return results

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': expect_ValueError_already_in_use})


async def test_two_threads_concurrently(run, num_workers):
    def func(ret):
        time.sleep(0.01)
        return ret

    async def coro(ctx):
        results = await asyncio.gather(
            ctx.call_in_thread(func, 'FOO'),
            ctx.call_in_thread(func, 'BAR'),
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, BaseException):
                raise result
        return results

    job = TJob('foo', coro=coro)
    if num_workers < 2:  # Cannot reserve second worker
        with job.thread_xevents():  # Only one thread started
            expect_result = RuntimeError(
                f'Cannot reserve >={num_workers} worker(s)!'
            )
    else:
        with job.thread_xevents():
            with job.thread_xevents():  # Both threads started
                expect_result = ['FOO', 'BAR']
    done = await run([job])
    assert verify_tasks(done, {'foo': expect_result})


async def test_two_subprocesses_sequentially_on_same_ticket_succeeds(run):
    argv = mock_argv()

    async def coro(ctx):
        async with ctx.reserve_worker() as ticket:
            with ctx.tjob.subprocess_xevents(argv, result=0):
                result = await ctx.run_in_subprocess(argv, ticket=ticket)
            with ctx.tjob.subprocess_xevents(
                argv, result=0, reuse_ticket=True
            ):
                result += await ctx.run_in_subprocess(argv, ticket=ticket)
            return result

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': 0})


async def test_two_subprocesses_concurrently_on_same_ticket_fails(run):
    argv = mock_argv()

    async def coro(ctx):
        async with ctx.reserve_worker() as ticket:
            with ctx.tjob.subprocess_xevents(argv, result=0):  # Only one
                results = await asyncio.gather(
                    ctx.run_in_subprocess(argv, ticket=ticket),
                    ctx.run_in_subprocess(argv, ticket=ticket),
                    return_exceptions=True,
                )
            for result in results:
                if isinstance(result, BaseException):
                    raise result
            return results

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': expect_ValueError_already_in_use})


async def test_two_subprocesses_concurrently(run, num_workers):
    argv = mock_argv()

    async def coro(ctx):
        results = await asyncio.gather(
            ctx.run_in_subprocess(argv),
            ctx.run_in_subprocess(argv),
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, Exception):
                raise result
        return results

    job = TJob('foo', coro=coro)
    if num_workers < 2:  # Cannot reserve second worker
        with job.subprocess_xevents(argv, result=0):  # Only one subprocess
            expect_result = RuntimeError(
                f'Cannot reserve >={num_workers} worker(s)!'
            )
    else:
        with job.subprocess_xevents(argv, result=0):
            with job.subprocess_xevents(argv, result=0):
                expect_result = [0, 0]
    done = await run([job])
    assert verify_tasks(done, {'foo': expect_result})


async def test_two_subprocesses_with_output_concurrently(run, num_workers):
    argv1 = mock_argv('FOO')
    argv2 = mock_argv('BAR')

    async def coro(ctx):
        async with ctx.subprocess(argv1, stdout=PIPE) as proc1:
            async with ctx.subprocess(argv2, stdout=PIPE) as proc2:
                output2 = await proc2.stdout.read()
                await proc2.wait()
            output1 = await proc1.stdout.read()
            await proc1.wait()
        return output1 + output2

    job = TJob('foo', coro=coro)
    if num_workers < 2:  # Cannot reserve second worker
        with job.subprocess_xevents(argv1, result='terminate'):
            expect_result = RuntimeError(
                f'Cannot reserve >={num_workers} worker(s)!'
            )
    else:
        with job.subprocess_xevents(argv1, result=0):
            with job.subprocess_xevents(argv2, result=0):
                expect_result = b'FOO\nBAR\n'
    done = await run([job])
    assert verify_tasks(done, {'foo': expect_result})


async def test_pass_data_between_subprocesses(run, num_workers):
    if num_workers < 2:
        pytest.skip('need Scheduler with at least 2 workers')

    argv_src = mock_argv('FOO', 'BAR', 'BAZ')
    argv_dst = mock_argv('foo', 'in:', 'bar', 'in:', 'baz', 'in:')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv_src, result=0):
            async with ctx.subprocess(argv_src, stdout=PIPE) as proc_src:
                with ctx.tjob.subprocess_xevents(argv_dst, result=0):
                    async with ctx.subprocess(
                        argv_dst, stdin=PIPE, stdout=PIPE
                    ) as proc_dst:
                        await copy_lines(proc_src.stdout, proc_dst.stdin)
                        output = await proc_dst.stdout.read()
                        await proc_dst.wait()
                await proc_src.wait()
        return output

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': b'foo\nFOO\nbar\nBAR\nbaz\nBAZ\n'})


async def test_pass_data_between_subprocesses_outer_fails(run, num_workers):
    if num_workers < 2:
        pytest.skip('need Scheduler with at least 2 workers')

    argv_src = mock_argv('FOO', 'BAR', 'exit:1')
    argv_dst = mock_argv('foo', 'in:', 'bar', 'in:', 'baz', 'in:')
    # Outer/src process does exit 1 while inner process is waiting for input.
    # Inner/dst process gets EOF, prints empty string, and does exit 0.

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv_src, result=1):
            async with ctx.subprocess(
                argv_src, stdout=PIPE, check=True
            ) as proc_src:
                with ctx.tjob.subprocess_xevents(argv_dst, result=0):
                    async with ctx.subprocess(
                        argv_dst, stdin=PIPE, stdout=PIPE, check=True
                    ) as proc_dst:
                        await copy_lines(proc_src.stdout, proc_dst.stdin)
                        output = await proc_dst.stdout.read()
                        await proc_dst.wait()
                await proc_src.wait()
        return output

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': CalledProcessError(1, argv_src)})


async def test_pass_data_between_subprocesses_inner_fails(run, num_workers):
    if num_workers < 2:
        pytest.skip('need Scheduler with at least 2 workers')

    argv_src = mock_argv('FOO', 'BAR', 'BAZ')
    argv_dst = mock_argv('foo', 'in:', 'bar', 'in:', 'exit:2')
    # Outer/src process produces all its output and does exit 0.
    # Inner/dst process does exit 2 without consuming all input.
    # Adding check=True here complicates the test as the outer process
    # may or may not be terminated by the inner's CalledProcessError(2).

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv_src, result=0):
            async with ctx.subprocess(
                argv_src, stdout=PIPE, check=True
            ) as proc_src:
                with ctx.tjob.subprocess_xevents(argv_dst, result=2):
                    async with ctx.subprocess(
                        argv_dst, stdin=PIPE, stdout=PIPE
                    ) as proc_dst:
                        output = b''
                        await copy_lines(proc_src.stdout, proc_dst.stdin)
                        output += await proc_dst.stdout.read()
                        await proc_dst.wait()
                await proc_src.wait()
        return proc_dst.returncode, output

    job = TJob('foo', coro=coro)
    done = await run([job])
    assert verify_tasks(done, {'foo': (2, b'foo\nFOO\nbar\nBAR\n')})


# add_thread_job() helper


async def test_one_threaded_job_before_another(Scheduler):
    def first_job():
        return 'return value from first job'

    def second_job():
        return 'return value from second job'

    scheduler = Scheduler()
    scheduler.add_thread_job('first_job', first_job)
    scheduler.add_thread_job('second_job', second_job, deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done,
        {
            'first_job': 'return value from first job',
            'second_job': 'return value from second job',
        },
    )


async def test_one_threaded_failure_cancels_the_next(Scheduler):
    def first_job():
        raise ValueError('FAIL')

    def second_job():
        return 'return value from second job'

    scheduler = Scheduler()
    scheduler.add_thread_job('first_job', first_job)
    scheduler.add_thread_job('second_job', second_job, deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done, {'first_job': ValueError('FAIL'), 'second_job': Cancelled},
    )


# add_subprocess_job() helper


async def test_one_subproc_job_before_another(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['true'])
    scheduler.add_subprocess_job('second_job', ['false'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(done, {'first_job': 0, 'second_job': 1},)


async def test_one_subproc_job_before_another_with_check_fails(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['true'], check=True)
    scheduler.add_subprocess_job(
        'second_job', ['false'], deps={'first_job'}, check=True
    )
    done = await scheduler.run()

    assert verify_tasks(
        done, {'first_job': 0, 'second_job': CalledProcessError(1, ['false'])},
    )


async def test_one_subprocess_failure_with_check_cancels_the_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'], check=True)
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(
        done,
        {
            'first_job': CalledProcessError(1, ['false']),
            'second_job': Cancelled,
        },
    )


async def test_one_subprocess_failure_wo_check_proceeds_to_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'], check=False)
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(done, {'first_job': 1, 'second_job': 0})


async def test_one_subprocess_failure_by_default_proceeds_to_next(Scheduler):
    scheduler = Scheduler()
    scheduler.add_subprocess_job('first_job', ['false'])
    scheduler.add_subprocess_job('second_job', ['true'], deps={'first_job'})
    done = await scheduler.run()

    assert verify_tasks(done, {'first_job': 1, 'second_job': 0})



import asyncio
import pytest
from subprocess import PIPE
import time

from asyncjobs import external_work, signal_handling

from conftest import (
    abort_in,
    assert_elapsed_time,
    Cancelled,
    mock_argv,
    TExternalWorkJob,
    verified_events,
    verify_number_of_tasks,
    verify_tasks,
)

pytestmark = pytest.mark.asyncio

TJob = TExternalWorkJob


@pytest.fixture
def Scheduler(scheduler_with_workers):
    return scheduler_with_workers(
        signal_handling.Scheduler, external_work.Scheduler
    )


@pytest.fixture
def run(Scheduler):
    async def _run(todo, abort_after=None):
        scheduler = Scheduler()
        with verified_events(scheduler, todo):
            for job in todo:
                scheduler.add_job(job.name, job, getattr(job, 'deps', None))
            try:
                async with abort_in(abort_after):
                    return await scheduler.run()
            finally:
                verify_number_of_tasks(1)  # no other tasks than test itself

    return _run


# aborting jobs shall properly clean up all jobs + scheduler


async def test_return_before_abort(run):
    todo = [TJob('foo', async_sleep=0.1)]
    with assert_elapsed_time(lambda t: t < 0.4):
        done = await run(todo, abort_after=0.5)
    assert verify_tasks(done, {'foo': 'foo done'})


async def test_abort_one_job_returns_immediately(run):
    todo = [TJob('foo', async_sleep=5)]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_job_in_thread_cannot_return_immediately(run):
    todo = [TJob('foo', thread=lambda ctx: time.sleep(0.2))]
    with assert_elapsed_time(lambda t: t > 0.2):  # must wait for thread
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_job_in_subproc_returns_immediately(run):
    todo = [TJob('foo', argv=mock_argv('sleep:5'))]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_one_spawned_job_returns_immediately(run):
    todo = [TJob('foo', spawn=[TJob('bar', async_sleep=5)], await_spawn=True)]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(done, {'foo': Cancelled, 'bar': Cancelled})


async def test_abort_one_non_terminating_job_teminates_then_kills(run):
    argv = mock_argv('ignore:SIGTERM', 'ignore:SIGINT', 'FOO', 'sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='kill'):
            async with ctx.subprocess(
                argv, stdout=PIPE, kill_delay=0.1
            ) as proc:
                assert b'FOO\n' == await proc.stdout.readline()
                # Subprocess will now ignore SIGTERM when we are cancelled
                async with abort_in(0.1):
                    await proc.wait()

    todo = [TJob('foo', coro=coro)]
    with assert_elapsed_time(lambda t: t < 2):
        done = await run(todo)
    assert verify_tasks(done, {'foo': Cancelled})
    # assert False


async def test_abort_job_with_two_subprocs_terminates_both(run, num_workers):
    if num_workers < 2:
        pytest.skip('need Scheduler with at least 2 workers')

    argv = mock_argv('sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv, result='terminate'):
            async with ctx.subprocess(argv) as proc1:
                with ctx.tjob.subprocess_xevents(argv, result='terminate'):
                    async with ctx.subprocess(argv) as proc2:
                        await proc2.wait()
                await proc1.wait()

    todo = [TJob('foo', coro=coro)]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.2)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_job_with_two_non_terminating_kills_both(run, num_workers):
    if num_workers < 2:
        pytest.skip('need Scheduler with at least 2 workers')

    argv1 = mock_argv('ignore:SIGTERM', 'ignore:SIGINT', 'FOO', 'sleep:5')
    argv2 = mock_argv('ignore:SIGTERM', 'ignore:SIGINT', 'BAR', 'sleep:5')

    async def coro(ctx):
        with ctx.tjob.subprocess_xevents(argv1, result='kill'):
            async with ctx.subprocess(
                argv1, stdout=PIPE, kill_delay=0.1
            ) as proc1:
                assert b'FOO\n' == await proc1.stdout.readline()
                with ctx.tjob.subprocess_xevents(argv2, result='kill'):
                    async with ctx.subprocess(
                        argv2, stdout=PIPE, kill_delay=0.1
                    ) as proc2:
                        assert b'BAR\n' == await proc2.stdout.readline()
                        # Both subprocesses will now ignore SIGTERM and we can
                        # proceed with cancelling.
                        async with abort_in(0.1):
                            await proc2.wait()
                await proc1.wait()

    todo = [TJob('foo', coro=coro)]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo)
    assert verify_tasks(done, {'foo': Cancelled})


async def test_abort_many_jobs_returns_immediately(num_jobs, run):
    todo = [TJob(f'foo #{i}', async_sleep=5) for i in range(num_jobs)]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(
        done, {f'foo #{i}': Cancelled for i in range(num_jobs)}
    )


async def test_abort_many_jobs_in_threads_cannot_return_soon(num_jobs, run):
    todo = [
        TJob(f'foo #{i}', thread=lambda ctx: time.sleep(0.3))
        for i in range(num_jobs)
    ]
    with assert_elapsed_time(lambda t: t > 0.3):  # must wait for all threads
        done = await run(todo, abort_after=0.1)

    # In some scenarios it seems a few jobs manage to successfully complete
    # before we get around to aborting the rest.
    def cancelled_xor_successful(task):
        try:
            if task.result() is None:
                return True
        except asyncio.CancelledError:
            return True
        except BaseException:
            pass
        return False

    assert verify_tasks(
        done, {f'foo #{i}': cancelled_xor_successful for i in range(num_jobs)}
    )


async def test_abort_many_jobs_in_subprocs_returns_immediately(num_jobs, run):
    todo = [
        TJob(f'foo #{i}', argv=mock_argv('sleep:30')) for i in range(num_jobs)
    ]
    with assert_elapsed_time(lambda t: t < 10.0):
        done = await run(todo, abort_after=0.1)
    assert verify_tasks(
        done, {f'foo #{i}': Cancelled for i in range(num_jobs)}
    )


async def test_abort_many_spawned_jobs_returns_immediately(num_jobs, run):
    todo = [
        TJob(
            'foo',
            spawn=[TJob(f'bar #{i}', async_sleep=5) for i in range(100)],
            await_spawn=True,
        )
    ]
    with assert_elapsed_time(lambda t: t < 1):
        done = await run(todo, abort_after=0.1)
    expect = {f'bar #{i}': Cancelled for i in range(100)}
    expect['foo'] = Cancelled
    assert verify_tasks(done, expect)
